"""
Sets up a general Interpolator class that looks like Mathematica's InterpolatingFunction class
"""

import numpy as np, abc, enum, math
import scipy.interpolate as interpolate
import scipy.spatial as spat
from .Mesh import Mesh, MeshType

from .NeighborBasedInterpolators import *

__all__ = [
    "Interpolator",
    "Extrapolator",
    "RBFDInterpolator",
    "InverseDistanceWeightedInterpolator",
    "ProductGridInterpolator",
    "UnstructuredGridInterpolator"
]


class InterpolatorException(Exception):
    pass


######################################################################################################
##
##                                   Interpolator Class
##
######################################################################################################

class BasicInterpolator(metaclass=abc.ABCMeta):
    """
    Defines the abstract interface we'll need interpolator instances to satisfy so that we can use
    `Interpolator` as a calling layer
    """

    @abc.abstractmethod
    def __init__(self, grid, values, **opts):
        raise NotImplementedError("abstract interface")

    @abc.abstractmethod
    def __call__(self, points, **kwargs):
        raise NotImplementedError("abstract interface")

    @abc.abstractmethod
    def derivative(self, order):
        """
        Constructs the derivatives of the interpolator at the given order
        :param order:
        :type order:
        :return:
        :rtype: BasicInterpolator
        """
        raise NotImplementedError("abstract interface")

class ProductGridInterpolator(BasicInterpolator):
    """
    A set of interpolators that support interpolation
    on a regular (tensor product) grid
    """

    def __init__(self, grids, vals, caller=None, order=None, extrapolate=True):
        """

        :param grids:
        :type grids:
        :param points:
        :type points:
        :param caller:
        :type caller:
        :param order:
        :type order: int | Iterable[int]
        """

        if order is None:
            order = 3

        self.grids = grids
        self.vals = vals
        if caller is None:
            if isinstance(grids[0], (int, float, np.integer, np.floating)):
                grids = [grids]
            ndim = len(grids)
            if ndim == 1:
                opts = {}
                if order is not None:
                    opts["k"] = order

                caller = interpolate.PPoly.from_spline(interpolate.splrep(grids[0], vals, k=order),
                                                      extrapolate=extrapolate
                                                      )
            else:
                caller = self.construct_ndspline(grids, vals, order, extrapolate=extrapolate)
        self.caller = caller

    @classmethod
    def construct_ndspline(cls, grids, vals, order, extrapolate=True):
        """
        Builds a tensor product ndspline by constructing a product of 1D splines

        :param grids: grids for each dimension independently
        :type grids: Iterable[np.ndarray]
        :param vals:
        :type vals: np.ndarray
        :param order:
        :type order: int | Iterable[int]
        :return:
        :rtype: interpolate.NdPPoly
        """

        # inspired by `csaps` python package
        # idea is to build a spline approximation in
        # every direction (where the return values are multidimensional)
        ndim = len(grids)
        if isinstance(order, (int, np.integer)):
            order = [order] * ndim

        coeffs = vals
        x = [None]*ndim
        for i, (g, o) in enumerate(zip(grids, order)):
            og_shape = coeffs.shape
            coeffs = coeffs.reshape((len(g), -1)).T
            sub_coeffs = [np.empty(0)]*len(coeffs)
            for e,v in enumerate(coeffs):
                ppoly = interpolate.PPoly.from_spline(interpolate.splrep(g, v, k=o))
                x[i] = ppoly.x
                sub_coeffs[e] = ppoly.c
            coeffs = np.array(sub_coeffs)
            coeffs = coeffs.reshape(
                og_shape[1:]+
                    sub_coeffs[0].shape
            )
            tot_dim = ndim+i+1
            coeffs = np.moveaxis(coeffs, tot_dim-2, tot_dim-2-i)

        return interpolate.NdPPoly(coeffs, x, extrapolate=extrapolate)

    def __call__(self, *args, **kwargs):
        """
        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype: np.ndarray
        """
        return self.caller(*args, **kwargs)

    def derivative(self, order):
        """
        :param order:
        :type order:
        :return:
        :rtype: ProductGridInterpolator
        """
        # ndim = len(self.grids)
        # if ndim == 1:
        return type(self)(
            self.grids,
            self.vals,
            caller=self.caller.derivative(order)
        )
        # elif ndim == 2:
        #     def caller(coords, _og=self.caller, **kwargs):
        #         return _og(coords, dx=order[0], dy=order[1], **kwargs)
        #     return type(self)(
        #         self.grids,
        #         self.vals,
        #         caller=caller
        #     )
        # else:
        #     derivs = self.caller.derivative(order)
        #     raise NotImplementedError("woof")

class UnstructuredGridInterpolator(BasicInterpolator):
    """
    Defines an interpolator appropriate for totally unstructured grids by
    delegating to the scipy `RBF` interpolators
    """

    default_neighbors=25
    def __init__(self, grid, values, order=None, neighbors=None, extrapolate=True, **opts):
        """
        :param grid:
        :type grid: np.ndarray
        :param values:
        :type values:  np.ndarray
        :param order:
        :type order: int
        :param neighbors:
        :type neighbors: int
        :param extrapolate:
        :type extrapolate: bool
        :param opts:
        :type opts:
        """
        self.extrapolate=extrapolate
        self._hull = None
        self._grid = grid

        if neighbors is None:
            neighbors = np.min([self.default_neighbors, len(grid)])

        if order is not None:
            if isinstance(order, int):
                if order == 1:
                    order = "linear"
                elif order == 3:
                    order = "cubic"
                elif order == 5:
                    order = "quintic"
                else:
                    raise InterpolatorException("{} doesn't support interpolation order '{}'".format(
                        interpolate.RBFInterpolator,
                        order
                    ))
            self.caller = interpolate.RBFInterpolator(grid, values, kernel=order, neighbors=neighbors, **opts)
        else:
            self.caller = interpolate.RBFInterpolator(grid, values, neighbors=neighbors, **opts)

    def _member_q(self, points):
        """
        Checks if the points are in the interpolated convex hull
        in the case that we aren't extrpolating so we can return
        NaN for those points
        :param points:
        :type points:
        :return:
        :rtype:
        """
        if self._hull is None:
            self._hull = spat.ConvexHull(self._grid)
            self._hull = spat.Delaunay(self._hull)
        return self._hull.find_simplex(points) >= 0
    def __call__(self, points):
        if self.extrapolate:
            return self.caller(points)
        else:
            hull_points = self._member_q(points)
            res = np.full(len(points), np.nan)
            res[hull_points] = self.caller(points[hull_points])
            return res

    def derivative(self, order):
        """
        Constructs the derivatives of the interpolator at the given order
        :param order:
        :type order:
        :return:
        :rtype: UnstructuredGridInterpolator
        """
        raise NotImplementedError("derivatives not implemented for unstructured grids")

# class _PolyFitManager:
#     """
#     Provides ways to evaluate polynomials
#     fast and get matrices of terms to fit
#     """
#     def __init__(self, order, ndim):
#         self.order = order
#         self.ndim = ndim
#         self._expr = None
#     @property
#     def perms(self):
#         from ..Combinatorics import SymmetricGroupGenerator
#
#         if self._perms is None:
#             # move this off to cached classmethod
#             # since there are restricted numbers of (order, ndim) pairs
#             sn = SymmetricGroupGenerator(self.ndim)
#             self._perms = sn.get_terms(list(range(self.order+1)))
#         return self._perms
#     def eval_poly_mat(self, coords):
#         """
#
#         :param coords: shape=(npts, ndim)
#         :type coords: np.ndarray
#         :return:
#         :rtype:
#         """
#         return np.power(coords[:, np.newaxis, :], self.perms[np.newaxis])
#     def eval_poly_mat_deriv(self, which):
#         which, counts = np.unique(which, return_counts=True)
#         perm_inds = self.perms[:, which]
#         shift_perms = perm_inds - counts[np.newaxis, :] # new monom orders


class ExtrapolatorType(enum.Enum):
    Default='Automatic'
    Error='Raise'

class Interpolator:
    """
    A general purpose that takes your data and just interpolates it without whining or making you do a pile of extra work
    """
    DefaultExtrapolator = ExtrapolatorType.Default
    def __init__(self,
                 grid,
                 vals,
                 interpolation_function=None,
                 interpolation_order=None,
                 extrapolator=None,
                 extrapolation_order=None,
                 **interpolation_opts
                 ):
        """
        :param grid: an unstructured grid of points **or** a structured grid of points **or** a 1D array
        :type grid: np.ndarray
        :param vals: the values at the grid points
        :type vals: np.ndarray
        :param interpolation_function: the basic function to be used to handle the raw interpolation
        :type interpolation_function: None | BasicInterpolator
        :param interpolation_order: the order of extrapolation to use (when applicable)
        :type interpolation_order: int | str | None
        :param extrapolator: the extrapolator to use for data points not on the grid
        :type extrapolator: Extrapolator | None | str | function
        :param extrapolation_order: the order of extrapolation to use by default
        :type extrapolation_order: int | str | None
        :param interpolation_opts: the options to be fed into the interpolating_function
        :type interpolation_opts:
        """
        self.grid = grid = Mesh(grid) if not isinstance(grid, Mesh) else grid
        self.vals = vals
        if interpolation_function is None:
            interpolation_function = self.get_interpolator(grid, vals,
                                                           interpolation_order=interpolation_order,
                                                           allow_extrapolation=extrapolator is None,
                                                           **interpolation_opts
                                                           )
        self.interpolator = interpolation_function

        if extrapolator is not None:
            if isinstance(extrapolator, ExtrapolatorType):
                if extrapolator == ExtrapolatorType.Default:
                    extrapolator = self.get_extrapolator(grid, vals, extrapolation_order=extrapolation_order)
                else:
                    raise ValueError("don't know what do with extrapolator type {}".format(extrapolator))
            elif not isinstance(extrapolator, Extrapolator):
                extrapolator = Extrapolator(extrapolator)
        self.extrapolator = extrapolator

    @classmethod
    def get_interpolator(cls, grid, vals, interpolation_order=None, allow_extrapolation=True, **opts):
        """Returns a function that can be called on grid points to interpolate them

        :param grid:
        :type grid: Mesh
        :param vals:
        :type vals: np.ndarray
        :param interpolation_order:
        :type interpolation_order: int | str | None
        :param opts:
        :type opts:
        :return: interpolator
        :rtype: function
        """
        if grid.ndim == 1:
            interpolator = ProductGridInterpolator(
                grid,
                vals,
                order=interpolation_order,
                extrapolate=allow_extrapolation
            )
        elif (
                grid.mesh_type == MeshType.Structured
                or grid.mesh_type == MeshType.Regular
        ):
            interpolator = ProductGridInterpolator(
                grid.subgrids,
                vals,
                order=interpolation_order,
                extrapolate=allow_extrapolation
            )
        elif grid.mesh_type == MeshType.Unstructured:
            # for now we'll only use the RadialBasisFunction interpolator, but this may be extended in the future
            interpolator = UnstructuredGridInterpolator(
                grid,
                vals,
                order=interpolation_order,
                extrapolate=allow_extrapolation
            )
        elif grid.mesh_type == MeshType.SemiStructured:
            raise NotImplementedError("don't know what I want to do with semistructured meshes anymore")
        else:
            raise InterpolatorException("{}.{}: can't handle mesh_type '{}'".format(
                cls.__name__,
               'get_interpolator',
                grid.mesh_type
            ))

        return interpolator

    @classmethod
    def get_extrapolator(cls, grid, vals, extrapolation_order=1, **opts):
        """
        Returns an Extrapolator that can be called on grid points to extrapolate them

        :param grid:
        :type grid: Mesh
        :param extrapolation_order:
        :type extrapolation_order: int
        :return: extrapolator
        :rtype: Extrapolator
        """

        # Extrapolator(
        #     cls(
        #         grid,
        #         vals,
        #         interpolation_order=extrapolation_order,
        #         extrapolator=Extrapolator(lambda g: np.full(g.shape, np.nan))
        #     )
        # )

        if grid.ndim == 1:
            extrapolator = ProductGridInterpolator(
                grid,
                vals,
                order=extrapolation_order,
                extrapolate=True
            )
        elif (
                grid.mesh_type == MeshType.Structured
                or grid.mesh_type == MeshType.Regular
        ):
            extrapolator = ProductGridInterpolator(
                grid.subgrids,
                vals,
                order=extrapolation_order,
                extrapolate=True
            )
        elif grid.mesh_type == MeshType.Unstructured:
            # for now we'll only use the RadialBasisFunction interpolator, but this may be extended in the future
            extrapolator = UnstructuredGridInterpolator(
                grid,
                vals,
                neighbors=extrapolation_order+1,
                order=extrapolation_order,
                extrapolate=True
            )
        elif grid.mesh_type == MeshType.SemiStructured:
            raise NotImplementedError("don't know what I want to do with semistructured meshes anymore")
        else:
            raise InterpolatorException("{}.{}: can't handle mesh_type '{}'".format(
                cls.__name__,
               'get_interpolator',
                grid.mesh_type
            ))

        return Extrapolator(
            extrapolator,
            **opts
        )

    def apply(self, grid_points, **opts):
        """Interpolates then extrapolates the function at the grid_points

        :param grid_points:
        :type grid_points:
        :return:
        :rtype:
        """
        # determining what points are "inside" an interpolator region is quite tough
        # instead it is probably better to allow the basic thing to interpolate and do its thing
        # and then allow the extrapolator to post-process that result
        vals = self.interpolator(grid_points, **opts)
        if self.extrapolator is not None:
            vals = self.extrapolator(grid_points, vals)
        return vals

    def derivative(self, order):
        """
        Returns a new function representing the requested derivative
        of the current interpolator

        :param order:
        :type order:
        :return:
        :rtype:
        """
        return type(self)(
            self.grid,
            self.vals,
            interpolation_function=self.interpolator.derivative(order),
            extrapolator=self.extrapolator.derivative(order) if self.extrapolator is not None else self.extrapolator
        )

    def __call__(self, *args, **kwargs):
        return self.apply(*args, **kwargs)

######################################################################################################
##
##                                   Extrapolator Class
##
######################################################################################################
class Extrapolator:
    """
    A general purpose that takes your data and just extrapolates it.
    This currently only exists in template format.
    """
    def __init__(self,
                 extrapolation_function,
                 warning=False,
                 **opts
                 ):
        """
        :param extrapolation_function: the function to handle extrapolation off the interpolation grid
        :type extrapolation_function: None | function | Callable | Interpolator
        :param warning: whether to emit a message warning about extrapolation occurring
        :type warning: bool
        :param opts: the options to feed into the extrapolator call
        :type opts:
        """
        self.extrapolator = extrapolation_function
        self.extrap_warning = warning
        self.opts = opts

    def derivative(self, n):
        return type(self)(
            self.extrapolator.derivative(n),
            warning=self.extrapolator
        )

    def find_extrapolated_points(self, gps, vals, extrap_value=np.nan):
        """
        Currently super rough heuristics to determine at which points we need to extrapolate
        :param gps:
        :type gps:
        :param vals:
        :type vals:
        :return:
        :rtype:
        """
        if extrap_value is np.nan:
            where = np.isnan(vals)
        elif extrap_value is np.inf:
            where = np.isinf(vals)
        elif not isinstance(extrap_value, (int, float, np.floating, np.integer)):
            where = np.logical_and(vals <= extrap_value[0], vals >= extrap_value[1])
        else:
            where = np.where(vals == extrap_value)

        return gps[where], where

    def apply(self, gps, vals, extrap_value=np.nan):
        ext_gps, inds = self.find_extrapolated_points(gps, vals, extrap_value=extrap_value)
        if len(ext_gps) > 0:
            new_vals = self.extrapolator(ext_gps)
            # TODO: emit a warning about extrapolating if we're doing so?
            vals[inds] = new_vals
        return vals

    def __call__(self, *args, **kwargs):
        return self.apply(*args, **kwargs)


import json
import os.path

from .Jobs import ElectronicStructureJob, OptionsBlock, SystemBlock

__all__ = [
    "GaussianJob"
]

class GaussianOptionsBlock(OptionsBlock):
    opts_key = None

    job_params_json = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Templates', 'gaussian_opts.json')
    _json = None
    @classmethod
    def load_json(cls):
        if cls._json is None:
            with open(cls.job_params_json) as opts_json:
                cls._json = json.load(opts_json)
        return cls._json

    @classmethod
    def get_props(cls):
        return list(cls.load_json()[cls.opts_key].keys())

    @classmethod
    def check_subopts(cls, key, opt_list, opt_dict=None, ignore_missing=False):
        if opt_dict is None:
            opt_list, opt_dict = cls.prep_opts(opt_list)

        base_opts = cls.load_json()
        if ignore_missing and key not in base_opts:
            return

        valid_opts = set(k.lower() for k in base_opts[key])
        bad_opts = []
        for o in opt_list:
            if o.lower() not in valid_opts:
                bad_opts.append(o)
        for o in opt_dict:
            if o.lower() not in valid_opts:
                bad_opts.append(o)
        if len(bad_opts) > 0:
            raise ValueError("got unknown options {} for property {} (known values are {})")

    @classmethod
    def format_opts(cls, opt_list, opt_dict=None, wrap=False):
        if opt_dict is None:
            if isinstance(opt_list, str):
                return opt_list
            else:
                opt_list, opt_dict = cls.prep_opts(opt_list)

        opt_strings = [
            cls.format_opts(o)
            for o in opt_list
        ] + [
            "{}={}".format(k, cls.format_opts(o))
            for k, o in opt_dict.items()
        ]
        if wrap and len(opt_strings) > 0 or len(opt_strings) > 1:
            return "(" + ",".join(opt_strings) + ")"
        else:
            return "".join(opt_strings)

    def format_base_params(self, opts=None):
        base_params = []
        if opts is None:
            opts = self.opts
        for k,o in opts.items():
            if o is True:
                base_params.append(k)
            elif o is False:
                ...
            else:
                fmt = self.format_opts(o)
                if len(fmt) == 0:
                    base_params.append(k)
                else:
                    base_params.append(k+"="+fmt)
        return base_params

class GaussianLinkBlock(GaussianOptionsBlock):
    opts_key = "Link0"

    def get_params(self):
        base = self.format_base_params()
        return {'link0':"\n".join("%"+b for b in base)}

class GaussianLOTBlock(GaussianOptionsBlock):
    opts_key = "LevelOfTheory"

    @classmethod
    def get_props(cls):
        return ["basis_set", "level_of_theory"] + super().get_props()

    @classmethod
    def get_basis_set_map(cls):
        return {
            k.lower():k
            for k in cls.load_json()["BasisSets"]
        }

    def get_params(self):
        lot = self.opts.get("level_of_theory", self.opts)
        if isinstance(lot, str):
            lot = {lot:[]}
        else:
            lot = lot.copy()

        lots = []
        for method, opts in lot.items():
            opt_list, opt_dict = self.prep_opts(opts)

            if 'basis_set' in opt_dict:
                basis_set = opt_dict['basis_set']
                del opt_dict[basis_set]
            elif 'basis_set' in self.opts:
                basis_set = self.opts['basis_set']
            else:
                bs_map = self.get_basis_set_map()
                for k in opt_list:
                    k = k.lower()
                    if k in bs_map:
                        basis_set = bs_map[k]
                        break
                else:
                    basis_set = ""
                opt_list = [k for k in opt_list if k.lower() not in bs_map]

            self.check_subopts(method, opt_list, opt_dict, ignore_missing=True)

            bs_string = ("/"+basis_set) if len(basis_set) > 0 else ""
            lots.append(
                method + self.format_opts(opt_list, opt_dict, wrap=True) + bs_string
            )

        if len(lots) == 0:
            return {}
        else:
            return {'level_of_theory':"#" + " ".join(lots)}

class GaussianRouteBlock(GaussianOptionsBlock):
    opts_key = "Route"

    @property
    def special_param_dispatch(self):
        return {
            "freq":self.handle_freq
        }

    def handle_freq(self, opts):
        opt_list, opt_dict = self.prep_opts(opts)
        base = {}
        extra = {}
        for k, o in opt_dict.items():
            if k.lower() in {"selanharmonicmodes"}:
                opt_list.append(k)
                extra["select_anharmonic_modes"] = " ".join(str(i) for i in o)
            elif k.lower() in {"selnormalmodes"}:
                opt_list.append(k)
                extra["select_normal_modes"] = " ".join(str(i) for i in o)
            else:
                base[k] = o

        return (opt_list, base), extra

    linewidth = 80
    def get_params(self):
        base_params = {}
        extra_params = {}
        disp = self.special_param_dispatch
        for k,o in self.opts.items():
            d = disp.get(k.lower())
            if d is None:
                base_params[k] = o
            else:
                o1, o2 = d(o)
                base_params[k] = o1
                extra_params.update(o2)
        gp_params = self.format_base_params(base_params)
        blocks = [ [] ]
        bl = 1
        for g in gp_params:
            pad = len(g) + 1
            bl = bl + pad
            if bl > self.linewidth:
                blocks.append([])
                bl = 1 + pad
            blocks[-1].append(g)

        params = {}
        route_str = "\n".join("#" + " ".join(b) for b in blocks if len(b) > 0)
        if len(route_str) > 0:
            params['route'] = route_str
        params.update(extra_params)

        return params

class GaussianSystemBlock(SystemBlock):
    __props__ = SystemBlock.__props__ + ("variables", "constants")

    fmt_key = ""


    @classmethod
    def format_vars_block(cls, vars, float_fmt="{:11.8f}", joiner=None):
        if isinstance(vars, dict):
            if joiner is None:
                joiner = "="
            vars = vars.items()
        if joiner is None:
            joiner = " "
        return "\n".join(
            " {k}{joiner}{v}".format(
                k=k,
                joiner=joiner,
                v=v if isinstance(v, str) else float_fmt.format(v)
            )
            for k, v in vars
        )

    def format_coordinate_block(self):
        charge = self.opts.get("charge", 0)
        multiplicity = self.opts.get("multiplicity", 1)
        carts = self.opts.get("cartesians")
        zmat = self.opts.get("zmatrix")

        chunks = []
        chunks.append(f"{charge} {multiplicity}")

        if carts is not None:
            if isinstance(carts, str):
                chunks.append(carts)
            else:
                chunks.append(self.fmt_carts(self.opts.get('atoms'), carts))
        elif zmat is not None:
            if isinstance(zmat, str):
                chunks.append(zmat)
            else:
                chunks.append(self.fmt_zmat(self.opts.get('atoms'), zmat, self.opts.get('ordering')))
        else:
            raise ValueError("no coordinate spec supplied")

        consts = self.opts.get("constants")
        if consts is None: consts = []
        if len(consts) > 0:
            chunks.append("Constants:")
            chunks.append(self.format_vars_block(consts))
        vars = self.opts.get("variables")
        if vars is None: vars = []
        if len(vars) > 0:
            chunks.append("Variables:")
            chunks.append(self.format_vars_block(vars))

        return "\n".join(chunks)

    def get_params(self):
        base_opts = {}
        if len(self.opts) > 0:
            base_opts[self.fmt_key + "system"] = self.format_coordinate_block()
            if len(self.opts.get('bonds', [])) > 0:
                base_opts[self.fmt_key + "bonds"] = self.format_bonds_block()
        return base_opts

class GaussianRestBlock(GaussianOptionsBlock):
    opts_key = "rest"
    job_template = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Templates', 'gaussian_job.gjf')
    @classmethod
    def load_json(cls):
        if cls._json is None:
            with open(cls.job_template) as r:
                t = r.read()
            return {"rest":{s.strip("{").strip("}"):[] for s in t.split()}}
        return cls._json

class GaussianJob(ElectronicStructureJob):
    job_template = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Templates', 'gaussian_job.gjf')
    blocks = [
        GaussianLinkBlock,
        GaussianLOTBlock,
        GaussianRouteBlock,
        GaussianSystemBlock,
        GaussianRestBlock
    ]

    def __init__(self, *strs, **opts):
        for o in strs:
            rt, o = GaussianRouteBlock.check_canon(o)
            if rt:
                opts[o] = True
            else:
                opts['level_of_theory'] = o
        super().__init__(**opts)

    @classmethod
    def get_extra_keys(cls):
        with open(cls.job_template) as r:
            t = r.read()
        return {s.strip("{").strip("}") for s in t.split()}

    @classmethod
    def get_block_types(cls):
        return cls.blocks

    @classmethod
    def load_template(cls):
        return cls.job_template

    non_blank_line_terminated = {'link0', 'level_of_theory'}
    def get_params(self):
        base_params = super().get_params()
        for k,b in base_params.items():
            if k not in self.non_blank_line_terminated:
                base_params[k] = b + "\n"

        return base_params



import uuid

from .JHTML import HTML

__all__ = [
    "JSMol"
]

__reload_hooks__ = [".JHTML"]

class JSMol:
    class Applet(HTML.Div):
        base_scripts = [
            HTML.Script(src="https://cdn.jsdelivr.net/gh/b3m2a1/jsmol-cdn@16.3.7.5/jsmol/JSmol.min.js"),
            HTML.Script(src="https://cdn.jsdelivr.net/gh/b3m2a1/jsmol-cdn@16.3.7.5/jsmol/js/Jmol2.js"),
            HTML.Script(
                """jmolInitialize("https://cdn.jsdelivr.net/gh/b3m2a1/jsmol-cdn@16.3.7.5/jsmol/");\n"""
                """Jmol.Info["serverURL"] = "https://chemapps.stolaf.edu/jmol/jsmol/php/jsmol.php";\n"""
                # """Jmol.Info["j2sPath"] = "https://cdn.jsdelivr.net/gh/b3m2a1/jsmol-cdn@16.3.7.4/jsmol/j2s";\n"""
                """Jmol.Info["width"] = "100%";\n"""
                """Jmol.Info["height"] = "100%";"""
            )
        ]
        def __init__(self, *model_etc, width='500px', height='500px', animate=False, vibrate=False, load_script=None,
                     suffix=None,
                     **attrs):
            if suffix is None:
                suffix = str(uuid.uuid4())[:6].replace("-", "")
            self.suffix = suffix
            self.id = "jsmol-applet-" + self.suffix
            if len(model_etc) > 0 and isinstance(model_etc[0], str):
                model_file = model_etc[0]
                rest = model_etc[1:]
            else:
                model_file = None
                rest = model_etc
            if load_script is None:
                if animate:
                    load_script = 'anim mode palindrome; anim on;'
                elif vibrate:
                    load_script = "vibration on"
                else:
                    load_script = ""
            self.load_script = load_script
            elems = self.base_scripts + [self.create_applet(model_file)] + list(rest)
            super().__init__(*elems, id=self.id, width=width, height=height, **attrs)

        @property
        def applet_target(self):
            return f"_{self.suffix}"
        def create_applet(self, model_file):
            targ = self.applet_target
            if model_file is None:
                loader = f'jmolApplet(400, "load {model_file}; {self.load_script}", "{targ}")'
            elif (
                    model_file.startswith("https://")
                    or model_file.startswith("file://")
                    or model_file.startswith("http://")
            ):
                loader = f'jmolApplet(400, "load {model_file}; {self.load_script}", "{targ}")'
            else:
                loader = f'jmolAppletInline(400, `{model_file}`, "{self.load_script}", "{targ}")'
            return HTML.Div(
                HTML.Script(f"""window.setTimeout((function () {{
                    let loaded = false;
                    if (!loaded) {{
                        loaded = true;
                        let applet = {loader};
                         document.getElementById("{self.id}").innerHTML = applet._code;
                        // window.setTimeout(() => {{ applet._cover(false); }}, 250)
                    }}
                }}), 250)"""),
                id=f"{self.id}-pane"
            )



"""
Provides a uniform interface for generating layouts from templates
by walking an object and its properties/children.
"""

__all__ = []
from .ObjectWalker import *; from .ObjectWalker import __all__ as exposed
__all__ += exposed
from .TemplateEngine import *; from .TemplateEngine import __all__ as exposed
__all__ += exposed


"""
Miscellaneous tools for interactive messing around in Jupyter environments
"""
import sys, os, types, importlib, inspect

__all__ = [
    "ModuleReloader",
    "ExamplesManager",
    "NotebookExporter",
    "FormattedTable",
    "NoLineWrapFormatter",
    "patch_pinfo"
]

__reload_hook__ = ['.NBExporter']

class ModuleReloader:
    """
    Reloads a module & recursively descends its 'all' tree
    to make sure that all submodules are also reloaded
    """

    def __init__(self, modspec):
        """
        :param modspec:
        :type modspec: str | types.ModuleType
        """
        if isinstance(modspec, str):
            modspec=sys.modules[modspec]
        self.mod = modspec

    def get_parents(self):
        """
        Returns module parents
        :return:
        :rtype:
        """
        split = self.mod.__name__.split(".")
        return [".".join(split[:i]) for i in range(len(split)-1, 0, -1)]

    def get_members(self):
        """
        Returns module members
        :return:
        :rtype:
        """

        base = self.mod.__all__ if hasattr(self.mod, '__all__') else dir(self.mod)
        if hasattr(self.mod, '__reload_hook__'):
            try:
                others = list(self.mod.__reload_hook__)
            except TypeError:
                pass
            else:
                base = [list(base), others]
        return base

    def reload_member(self, member,
        stack=None,
        reloaded=None, blacklist=None, reload_parents=True,
        verbose=False,
        print_indent=""
        ):

        # print(print_indent + " member:", member)
        if member.startswith('.'):
            how_many = 0
            while member[how_many] == ".":
                how_many += 1
                if how_many == len(member):
                    break
            main_name = self.mod.__name__.rsplit(".", how_many)[0]
            test_key = main_name + "." + member[how_many:]
        else:
            test_key = self.mod.__name__ + "." + member
        if test_key in sys.modules:
            type(self)(test_key).reload(
                reloaded=reloaded, blacklist=blacklist,
                verbose=verbose,
                reload_parents=reload_parents, print_indent=print_indent
            )
        else:
            obj = getattr(self.mod, member)
            if isinstance(obj, types.ModuleType):
                type(self)(obj).reload(
                    reloaded=reloaded, blacklist=blacklist,
                    verbose=verbose,
                    reload_parents=reload_parents, print_indent=print_indent
                )
            elif isinstance(obj, (type, types.MethodType, types.FunctionType)):
                type(self)(obj.__module__).reload(
                    reloaded=reloaded, blacklist=blacklist,
                    verbose=verbose,
                    reload_parents=reload_parents, print_indent=print_indent
                )
            else:
                # try:
                #     isinstance(obj, (type, types.FunctionType))
                # except Exception as e:
                #     print(e)
                # else:
                #     print("...things can be functions")
                obj = type(obj)
                type(self)(obj.__module__).reload(
                    reloaded=reloaded, blacklist=blacklist,
                    verbose=verbose,
                    reload_parents=reload_parents, print_indent=print_indent
                )

    blacklist_keys = ['site-packages', os.path.abspath(os.path.dirname(inspect.getfile(os)))]
    def reload(self, 
        stack=None,
        reloaded=None, blacklist=None, reload_parents=True, 
        verbose=False,
        print_indent=""
        ):
        """
        Recursively searches for modules to reload and then reloads them.
        Uses a cache to break cyclic dependencies of any sort.
        This turns out to also be a challenging problem, since we need to basically
        load depth-first, while never jumping too far back...


        :return:
        :rtype:
        """

        if reloaded is None:
            reloaded = set()

        if blacklist is None:
            blacklist = set()
        blacklist.update(sys.builtin_module_names)

        key = self.mod.__name__
        if (
                key not in reloaded
                and key not in blacklist
                and all(k not in inspect.getfile(self.mod) for k in self.blacklist_keys)
        ):
            if verbose:
                print(print_indent + "Reloading:", self.mod.__name__)
            reloaded.add(self.mod.__name__)

            print_indent += "  "

            mems = self.get_members()

            if isinstance(mems[0], list):
                req, opts = mems
            else:
                req = mems
                opts = []

            for member in req:
                self.reload_member(member,
                                   stack=stack,
                                   reloaded=reloaded,
                                   blacklist=blacklist,
                                   reload_parents=reload_parents,
                                   verbose=verbose,
                                   print_indent=print_indent
                                   )
            for member in opts:
                try:
                    self.reload_member(member,
                                       stack=stack,
                                       reloaded=reloaded,
                                       blacklist=blacklist,
                                       reload_parents=reload_parents,
                                       verbose=verbose,
                                       print_indent=print_indent
                                       )
                except:
                    pass

           
            
            if hasattr(self.mod, '__reload_hook__'):
                try:
                    self.mod.__reload_hook__()
                except TypeError:
                    pass
            if verbose:
                print(print_indent + "loading:", self.mod.__name__)
            importlib.reload(self.mod)


            load_parents = []
            if reload_parents:
                # make sure parents get loaded in the appropriate order...
                for parent in self.get_parents():
                    if parent in reloaded:
                        # prevent us from jumping back too far...
                        break
                    # print(" storing", parent)
                    load_parents.append(parent)
                    type(self)(parent).reload(
                        reloaded=reloaded, blacklist=blacklist, 
                        reload_parents=reload_parents, verbose=verbose,
                        print_indent=print_indent
                        )

    @classmethod
    def load_module(cls, module):
        if module in sys.modules:
            cls(module).reload()
        return importlib.import_module(module)

    @classmethod
    def import_from(cls, module, names, set_glob=True):
        mod = cls.load_module(module)
        objs = []
        single = isinstance(names, str)
        if single: names = [names]
        for name in names:
            obj = getattr(mod, name)
            if set_glob:
                globals()[name] = obj
            objs.append(obj)
        if single: objs = objs[0]
        return objs

class NotebookExporter:
    tag_filters = {
        'cell':('ignore',),
        'output':('ignore',),
        'input':('ignore',),
    }
    def __init__(self, name,
                 src_dir=None,
                 img_prefix=None,
                 img_dir=None,
                 output_dir=None,
                 tag_filters=None
                 ):
        self.name = name
        self.src_dir = src_dir
        self.out_dir = output_dir
        self.img_dir = img_dir
        self.img_prefix = img_prefix
        self.tag_filters = self.tag_filters if tag_filters is None else tag_filters

    def load_preprocessor(self):
        from .NBExporter import MarkdownImageExtractor
        prefix = '' if self.img_prefix is None else self.img_prefix

        return MarkdownImageExtractor(prefix=prefix)#lambda *args,prefix=prefix,**kw:print(args,kw)

    def load_filters(self):
        from traitlets.config import Config

        # Setup config
        c = Config()

        # Configure tag removal - be sure to tag your cells to remove  using the
        # words remove_cell to remove cells. You can also modify the code to use
        # a different tag word

        if 'cell' in self.tag_filters:
            c.TagRemovePreprocessor.remove_cell_tags = self.tag_filters['cell']
        if 'output' in self.tag_filters:
            c.TagRemovePreprocessor.remove_all_outputs_tags = self.tag_filters['output']
        if 'input' in self.tag_filters:
            c.TagRemovePreprocessor.remove_input_tags = self.tag_filters['input']
        c.TagRemovePreprocessor.enabled = True

        # Configure and run out exporter
        c.MarkdownExporter.preprocessors = [
            self.load_preprocessor(),
            "nbconvert.preprocessors.TagRemovePreprocessor"
        ]

        return c

    def load_nb(self):
        import nbformat

        this_nb = self.name + '.ipynb'
        if self.src_dir is not None:
            this_nb = os.path.join(self.src_dir, self.name+".ipynb")
        with open(this_nb) as nb:
            nb_cells = nbformat.reads(nb.read(), as_version=4)
        return nb_cells

    def save_output_file(self, filename, body):
        fn = os.path.abspath(filename)
        if fn != filename and self.img_dir is not None:
            filename = os.path.join(self.img_dir, filename)
        with open(filename, 'wb') as out:
            out.write(body)
        return filename

    def export(self):
        from nbconvert import MarkdownExporter
        nb_cells = self.load_nb()

        exporter = MarkdownExporter(config=self.load_filters())

        (body, resources) = exporter.from_notebook_node(nb_cells)

        # raise Exception(resources)
        if len(resources['outputs']) > 0:
            for k,v in resources['outputs'].items():
                self.save_output_file(k, v)

        out_md = self.name + '.md'
        if self.out_dir is not None:
            out_md = os.path.join(self.out_dir, self.name + ".md")

        with open(out_md, 'w+') as md:
            md.write(body)

        return out_md

# class DefaultExplorerInterface:
#     def ...
# class Explorer:
#     """
#     Provides a uniform interface for exploring what objects can do.
#     Hooks into the Jupyter runtime to provide nice interfaces
#     and has support for
#     """
#     def __init__(self, obj):
#         self.obj = obj
#
#     def _ipython_display_(self):
#         raise NotImplementedError("...")

class ExamplesManager:
    data_path = ("ci", "tests", "TestData")
    def __init__(self, root, data_path=None):
        if os.path.isdir(root):
            root = root
        else:
            if isinstance(root, str) and "/" not in root:
                root = importlib.import_module(root)
            if not isinstance(root, str):
                root = os.path.dirname(root.__file__)
        if data_path is None:
            data_path = self.data_path
        if isinstance(data_path, str):
            data_path = [data_path]
        self.test_dir = os.path.join(root, *data_path)

    def test_data(cls, *path):
        return os.path.join(cls.test_dir, *path)

    @classmethod
    def load_module(cls, module):
        return ModuleReloader.load_module(module)

    @classmethod
    def import_from(cls, module, names, set_glob=True):
        return ModuleReloader.import_from(module, names, set_glob=set_glob)

    @classmethod
    def parse_x3d_view_matrix(cls, vs, view_all=True):
        import json, numpy as np
        from .. import Numputils as nput

        # vs from JSON.stringify(document.getElementById('x3d').runtime.viewMatrix())
        vm = json.loads(vs)
        vm = np.linalg.inv([
            [vm[f"_{i}{j}"] for j in range(4)]
            for i in range(4)
        ])
        ang, ax = nput.extract_rotation_angle_axis(vm[:3, :3])
        v_pos = vm[:3, -1].tolist()
        v_ort = np.array(list(ax) + [ang]).tolist()
        opts = {"position": v_pos, "orientation": v_ort}
        if view_all:
            opts['viewAll'] = True
        return opts

class NoLineWrapFormatter:
    def __init__(self, *objs, white_space="pre", **opts):
        self.objs = [self._canonicalize(o) for o in objs]
        self.opts = dict(opts, white_space=white_space)
        self._widg = None
    def _canonicalize(self, o):
        if hasattr(o, 'to_widget') or hasattr(o, 'to_tree'):
            return o
        else:
            return str(o)
    def create_obj(self):
        from .JHTML import JHTML
        return JHTML.Pre(*self.objs, **self.opts)
    def to_widget(self):
        if self._widg is None:
            self._widg = self.create_obj()
        return self._widg
    # def show(self):
    #     return self.to_widget()
    def _ipython_display_(self):
        return self.to_widget()._ipython_display_()

class FormattedTable(NoLineWrapFormatter):
    def __init__(self, table_data, column_formats="8.3f", **format_opts):
        from ..Formatters import TableFormatter
        super().__init__(
            TableFormatter(column_formats, **format_opts).format(table_data)
        )


def patch_pinfo():
    from IPython.core.oinspect import Inspector
    from IPython.core.display import display

    if not hasattr(Inspector, '_og_pinfo'):
        Inspector._og_pinfo = Inspector.pinfo

    def pinfo(self, obj, oname='', formatter=None, info=None, detail_level=0, enable_html_pager=True):
        if hasattr(obj, '_ipython_pinfo_'):
            display(obj._ipython_pinfo_())
        else:
            return Inspector._og_pinfo(self, obj, oname=oname, formatter=formatter, info=info, detail_level=detail_level, enable_html_pager=enable_html_pager)

    Inspector.pinfo = pinfo



"""
Provides a general, convenient FiniteDifferenceFunction class to handle all of our difference FD imps
"""
import numpy as np, scipy.sparse as sparse, math
from ..Mesh import Mesh, MeshType
__reload_hook__ = [ '..Mesh' ]

__all__ = [
    'FiniteDifferenceFunction',
    'FiniteDifferenceError',
    'finite_difference',
    "FiniteDifference1D",
    "RegularGridFiniteDifference",
    "IrregularGridFiniteDifference",
    "FiniteDifferenceData",
    "FiniteDifferenceMatrix"
]

##########################################################################################
#
#                                    FiniteDifferenceFunction
#
##########################################################################################
class FiniteDifferenceError(Exception):
    pass
class FiniteDifferenceFunction:
    """
    The FiniteDifferenceFunction encapsulates a bunch of functionality extracted from [Fornberger's
    Calculation of Wieghts in Finite Difference Formulas](https://epubs.siam.org/doi/pdf/10.1137/S0036144596322507)

    Only applies to direct product grids, but each subgrid can be regular or irregular.
    Used in a large number of other places, but relatively rarely on its own.
    A convenient application is the `FiniteDifferenceDerivative` class in the `Derivatives` module.
    """
    def __init__(self, *diffs, axes=0, contract = False):
        """Constructs an object to take finite differences derivatives of grids of data

        :param diffs: A set of differences to take along successive axes in the data
        :type diffs: FiniteDifference1D
        :param axes: The axes to take the specified differences along
        :type axes: int | Iterable[int]
        :param contract: Whether to reduce the shape of the returned tensor if applicable after application
        :type contract: bool
        """
        self.differences = tuple(diffs)
        self.axes = axes
        self.contract = contract

    def apply(self, vals, axes=None, mesh_spacing = None, contract = None):
        """
        Iteratively applies the stored finite difference objects to the vals

        :param vals: The tensor of values to take the difference on
        :type vals: np.ndarray
        :param axes: The axis or axes to take the differences along (defaults to `self.axes`)
        :type axes: int | Iterable[int]
        :return: The tensor of derivatives
        :rtype: np.ndarray
        """
        dim = len(self.differences)
        val_dim = len(vals.shape)

        if contract is None:
            contract = self.contract

        axis = axes
        if axis is None:
            axis = self.axes
        if mesh_spacing is None or isinstance(mesh_spacing, (int, np.integer, float, np.floating)):
            mesh_spacing = [mesh_spacing] * dim

        # by default we'll assume dimension one is the slowest changing index, dimension 2 is the next slowest, etc.
        if isinstance(axis, (int, np.integer)):
            axis = axis + np.arange(dim)
        for i, m, h in zip(axis, self.differences, mesh_spacing):
            if m is not None:
                vals = m.apply(vals, val_dim=val_dim, axis=i, mesh_spacing=h)

        if contract:
            vals = vals.squeeze()
        return vals

    def __call__(self, vals, axes=None, mesh_spacing=None):
        return self.apply(vals, axes=axes, mesh_spacing=mesh_spacing)

    @property
    def order(self):
        """
        :return: the order of the derivative requested
        :rtype: tuple[int]
        """
        return tuple(d.order for d in self.differences)

    @property
    def weights(self):
        """
        :return: the weights for the specified stencil
        :rtype: tuple[np.array[float]]
        """
        return tuple(d.weights for d in self.differences)

    @property
    def widths(self):
        """
        :return: the number of points in each dimension, left and right, for the specified stencil
        :rtype: tuple[(int, int)]
        """
        return tuple(d.widths for d in self.differences)

    @classmethod
    def regular_difference(cls,
                           order,
                           mesh_spacing=None,
                           accuracy=2,
                           stencil=None,
                           end_point_accuracy=2,
                           axes=0,
                           contract=True,
                           **kwargs
                           ):
        """
        Constructs a `FiniteDifferenceFunction` appropriate for a _regular grid_ with the given stencil

        :param order: the order of the derivative
        :type order: tuple[int]
        :param mesh_spacing: the spacing between grid points in the regular grid `h`
        :type mesh_spacing: None | float | tuple[float]
        :param accuracy: the accuracy of the derivative that we'll try to achieve as a power on `h`
        :type accuracy: None | int | tuple[int]
        :param stencil: the stencil to use for the derivative (overrides `accuracy`)
        :type stencil: None | int | tuple[int]
        :param end_point_accuracy: the amount of extra accuracy to use at the edges of the grid
        :type end_point_accuracy: None | int | tuple[int]
        :param axes: the axes of the passed array for the derivative to be applied along
        :type axes: None | int | tuple[int]
        :param contract: whether to eliminate any axes of size `1` from the results
        :type contract: bool
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        dim = len(order)
        if isinstance(stencil, (int, np.integer)) or stencil is None:
            stencil = [stencil]*dim
        if isinstance(accuracy, (int, np.integer)) or accuracy is None:
            accuracy = [accuracy]*dim
        if isinstance(end_point_accuracy, (int, np.integer)) or end_point_accuracy is None:
            end_point_accuracy = [end_point_accuracy] * dim
        if mesh_spacing is None or isinstance(mesh_spacing, (int, np.integer, float, np.floating)):
            mesh_spacing = [mesh_spacing] * dim

        return cls(
            *(
                RegularGridFiniteDifference(
                    d,
                    stencil=s,
                    accuracy=a,
                    end_point_accuracy=e,
                    mesh_spacing = h,
                    **kwargs
                    ) for d, s, a, e, h in zip(order, stencil, accuracy, end_point_accuracy, mesh_spacing)
            ),
            axes=axes,
            contract=contract
        )

    @classmethod
    def from_grid(cls,
                  grid, order,
                  accuracy=2,
                  stencil=None,
                  end_point_accuracy=2,
                  axes=0,
                  contract=True,
                  **kwargs
                  ):
        """
        Constructs a `FiniteDifferenceFunction` from a grid and order.
         Deconstructs the grid into its subgrids and builds a different differencer for each dimension

        :param grid: The grid to use as input data when defining the derivative
        :type grid: np.ndarray
        :param order: order of the derivative to compute
        :type order: int or list of ints
        :param stencil: number of points to use in the stencil
        :type stencil: int or list of ints
        :return: deriv func
        :rtype: FiniteDifferenceFunction
        """

        if isinstance(order, (int, np.integer)):
            order = (order,)
        dim = len(order)
        if isinstance(stencil, (int, np.integer)) or stencil is None:
            stencil = [stencil]*dim
        if isinstance(accuracy, (int, np.integer)) or accuracy is None:
            accuracy = [accuracy]*dim
        if isinstance(end_point_accuracy, (int, np.integer)) or end_point_accuracy is None:
            end_point_accuracy = [end_point_accuracy] * dim

        if axes is None:
            axes = grid.ndim - dim - 1
        if isinstance(axes, (int, np.integer)):
            axes = axes + np.arange(dim)
        axes = tuple(axes)

        if dim > 1:
            if grid.ndim < dim+1:
                raise FiniteDifferenceError("{}.{}: grid has {} dimensions but {} dimensional difference requested".format(
                    cls.__name__,
                    "from_grid",
                    grid.ndim-1,
                    dim
                ))

            # get the grid we actually will be doing the FD on
            sel_ax = axes + (grid.ndim-1,)
            missing = tuple(ax for ax in range(grid.ndim) if ax not in sel_ax)
            transp_ax = missing+sel_ax
            new_grid = grid.transpose(transp_ax)
            if len(missing)>0:
                new_grid = new_grid[(0,)*len(missing)]

            gp = Mesh(new_grid)
            subgrids = gp.subgrids#[sg[a] if a is not None else a for a in sel_ax]
            if subgrids is None:
                raise ValueError("can't apply finite difference to mesh of type {}".format(gp.mesh_type))
        else:
            subgrids = [grid]

        diffs = [None] * dim
        for i, d in enumerate(zip(order, stencil, accuracy, end_point_accuracy, subgrids)):
            o, s, a, e, g = d
            if o is None or o == 0:
                diffs[i] = None
            else:
                if len(g) == 0:
                    raise ValueError("finite difference can't be applied to subgrid {} of {}".format(g, grid))

                m = Mesh(g)
                # raise Exception(..., m.mesh_type)
                if m.mesh_type == MeshType.Regular:
                    diffs[i] = RegularGridFiniteDifference(
                        o,
                        stencil=s,
                        accuracy=a,
                        end_point_accuracy=e,
                        mesh_spacing=g[1]-g[0],
                        **kwargs
                    )
                elif m.mesh_type is MeshType.Structured:
                    raise Exception(np.diff(g), np.unique(np.round(np.diff(g), 7)), np.unique(np.round(np.diff(subgrids[0]), 7)))
                    diffs[i] = IrregularGridFiniteDifference(
                        g,
                        o,
                        stencil=s,
                        accuracy=a,
                        end_point_accuracy=e,
                        **kwargs
                    )
                else:
                    raise ValueError("don't know how to do FD on a Mesh with type {}".format(m.mesh_type))
        return cls(*diffs, contract=contract, axes=axes)

##########################################################################################
#
#                                    FiniteDifference1D
#
##########################################################################################
class FiniteDifference1D:
    """
    A one-dimensional finite difference derivative object.
    Higher-dimensional derivatives are built by chaining these.
    """
    def __init__(self, finite_difference_data, matrix):
        self.data = finite_difference_data
        self.mat = matrix

    @property
    def order(self):
        return self.data.order
    @property
    def weights(self):
        return tuple(tuple(w) for w in self.data.weights)
    @property
    def widths(self):
        return tuple(tuple(w) for w in self.data.widths)

    @classmethod
    def get_stencil(cls, order, stencil, accuracy):
        if stencil is None:
            sten = order + accuracy - 1
        else:
            sten = stencil - 1
        return sten

    def apply(self, vals, val_dim=None, axis=0, mesh_spacing=None, check_shape=True):
        """
        Applies the held `FiniteDifferenceMatrix` to the array of values

        :param vals: values to do the difference over
        :type vals: np.ndarray | sparse.csr_matrix
        :param val_dim: dimensions of the vals
        :type val_dim: int
        :param axis: the axis to apply along
        :type axis: int | tuple[int]
        :param mesh_spacing: the mesh spacing for the weights
        :type mesh_spacing: float
        :return:
        :rtype: np.ndarray
        """

        if val_dim is None:
            val_dim = len(vals.shape)
        vs = vals.shape
        if val_dim == 1:
            # print(type(self.mat.matrix))
            if check_shape:
                if self.mat.only_core or self.mat.only_center:
                    shp = len(self.weights[len(self.weights) // 2])
                    if shp != vs[0]:
                        if self.mat.only_center:
                            raise FiniteDifferenceError(
                                "FD weights {} with can't be applied to data with shape {} along axis {}".format(
                                    self.weights[len(self.weights) // 2],
                                    self.weights,
                                    vs,
                                    axis
                                ))
                # else:
                #     shp = max(len(w) for w in self.weights)
                #     raise FiniteDifferenceError(
                #         "FD weights {} with len {} can't be applied to data with shape {} along axis {}".format(
                #             self.weights,
                #             shp,
                #             vs,
                #             axis
                #         ))
            if mesh_spacing is not None:
                self.mat.mesh_spacing = mesh_spacing
            self.mat.npts = vs[0]
            vals = self.mat.matrix.dot(vals)  # axis doesn't even make sense here...
            # print(vals.shape)
        else:
            if check_shape:
                if self.mat.only_core or self.mat.only_center:
                    shp = len(self.weights[ len(self.weights)//2 ])
                else:
                    shp = max(len(w) for w in self.weights)
                if shp != vs[axis]:
                    if self.mat.only_center:
                        raise FiniteDifferenceError(
                            "FD weights {} with can't be applied to data with shape {} along axis {}".format(
                                self.weights[len(self.weights) // 2],
                                # self.weights,
                                vs,
                                axis
                            ))
                    # else:
                    #     raise FiniteDifferenceError(
                    #         "FD weights {} with len {} can't be applied to data with shape {} along axis {}".format(
                    #             self.weights,
                    #             shp,
                    #             vs,
                    #             axis
                    #         ))
            if mesh_spacing is not None:
                self.mat.mesh_spacing = mesh_spacing
            self.mat.npts = vs[axis]
            if self.mat.mode == "sparse":
                vals = self.sparse_tensordot(self.mat.matrix, vals, axis)
            else:
                vals = np.tensordot(self.mat.matrix, vals, axes=((1), (axis)))
            roll = np.concatenate((
                np.roll(np.arange(axis + 1), -1),
                np.arange(axis + 1, val_dim)
            ))
            vals = vals.transpose(roll)

        return vals

    @staticmethod
    def sparse_tensordot(sparse, mat, axis):
        """Not sure how fast this will be, but does a very simple contraction of `mat` along `axis` by the final axis of `sparse`

        Heavily de-generalized from here: https://github.com/pydata/sparse/blob/9dc40e15a04eda8d8efff35dfc08950b4c07a810/sparse/_coo/common.py
        :param sparse:
        :type sparse: sparse.sparsemat
        :param mat:
        :type mat: np.ndarray
        :param axis:
        :type axis:
        :return:
        :rtype:
        """

        axis_a = 1
        axis_b = axis

        as_ = sparse.shape
        nda = sparse.ndim
        bs = mat.shape
        ndb = mat.ndim

        if as_[axis_a] != bs[axis_b]:
            raise ValueError("shape-mismatch for sum")
        if axis_b < 0:
            axis_b += ndb

        # Move the axes to sum over to the end of "a"
        # and to the front of "b"
        notin = [k for k in range(nda) if k != axis_a]
        newaxes_a = notin + [axis_a]
        N2 = 1
        N2 *= as_[axis_a]
        newshape_a = (-1, N2)
        olda = [as_[axis] for axis in notin]

        notin = [k for k in range(ndb) if k != axis_b]
        newaxes_b = [axis_b] + notin
        N2 = 1
        N2 *= bs[axis_b]
        newshape_b = (N2, -1)
        oldb = [bs[axis] for axis in notin]

        # don't think I need this...
        # at = sparse.transpose(newaxes_a).reshape(newshape_a)
        bt = mat.transpose(newaxes_b).reshape(newshape_b)
        res = np.asarray(sparse.dot(bt))
        return res.reshape(olda + oldb)

class RegularGridFiniteDifference(FiniteDifference1D):
    """
    Defines a 1D finite difference over a regular grid
    """
    def __init__(self,
                 order,
                 stencil = None,
                 accuracy = 4,
                 end_point_accuracy = 2,
                 **kw
                 ):
        """

        :param order: the order of the derivative to take
        :type order: int
        :param stencil: the number of stencil points to add
        :type stencil: int | None
        :param accuracy: the approximate accuracy to target with the method
        :type accuracy: int | None
        :param end_point_accuracy: the extra number of stencil points to add to the end points
        :type end_point_accuracy: int | None
        :param kw: options passed through to the `FiniteDifferenceMatrix`
        :type kw:
        """

        data = self.finite_difference_data(order, self.get_stencil(order, stencil, accuracy), end_point_accuracy)
        mat = FiniteDifferenceMatrix(data, **kw)
        super().__init__(data, mat)

    @classmethod
    def finite_difference_data(cls, order, stencil, end_point_precision):
        """Builds a FiniteDifferenceData object from an order, stencil, and end_point_precision

        :param order:
        :type order:
        :param stencil:
        :type stencil:
        :param end_point_precision:
        :type end_point_precision:
        :return:
        :rtype:
        """

        m = order
        outer_stencil = stencil + end_point_precision
        lefthand_coeffs = cls.get_weights(m, 0, outer_stencil)
        centered_coeffs = cls.get_weights(m, math.ceil(stencil / 2), stencil)
        righthand_coeffs = cls.get_weights(m, outer_stencil, outer_stencil)
        sten = len(centered_coeffs)
        widths = [
            [0, len(lefthand_coeffs)],
            [math.ceil(sten / 2), math.floor(sten / 2)],
            [len(righthand_coeffs), 0]
        ]
        coeffs = [lefthand_coeffs, centered_coeffs, righthand_coeffs]

        return FiniteDifferenceData(coeffs, widths, order)

    @staticmethod
    def get_weights(m, s, n):
        """Extracts the weights for an evenly spaced grid

        :param m:
        :type m:
        :param s:
        :type s:
        :param n:
        :type n:
        :return:
        :rtype:
        """
        from .ZachLib import EvenFiniteDifferenceWeights

        return EvenFiniteDifferenceWeights(m, s, n)

class IrregularGridFiniteDifference(FiniteDifference1D):
    """
    Defines a finite difference over an irregular grid
    """
    def __init__(self,
                 grid,
                 order,
                 stencil=None,
                 accuracy=2,
                 end_point_accuracy=2,
                 **kw
                 ):
        """

        :param grid: the grid to get the weights from
        :type grid: np.ndarray
        :param order: the order of the derivative to take
        :type order: int
        :param stencil: the number of stencil points to add
        :type stencil: int | None
        :param accuracy: the approximate accuracy to target with the method
        :type accuracy: int | None
        :param end_point_accuracy: the extra number of stencil points to add to the end points
        :type end_point_accuracy: int | None
        :param kw: options passed through to the `FiniteDifferenceMatrix`
        :type kw:
        """
        data = self.finite_difference_data(grid, order, self.get_stencil(order, stencil, accuracy), end_point_accuracy)
        mat = FiniteDifferenceMatrix(data, npts=len(grid), mesh_spacing=1, **kw)
        super().__init__(data, mat)

    @staticmethod
    def get_grid_slices(grid, stencil):
        """

        :param grid:
        :type grid:
        :param stencil:
        :type stencil:
        :return:
        :rtype:
        """
        return [grid[a:b] for a,b in zip(range(0, len(grid)-stencil), range(stencil, len(grid)))]

    @staticmethod
    def get_weights(m, z, x):
        """Extracts the grid weights for an unevenly spaced grid based off of the algorithm outlined by
        Fronberger in https://pdfs.semanticscholar.org/8bf5/912bde884f6bd4cfb4991ba3d077cace94c0.pdf

        :param m: highest derivative order
        :type m:
        :param z: center of the derivatives
        :type z:
        :param X: grid of points
        :type X:
        """
        from .ZachLib import UnevenFiniteDifferenceWeights  # loads from C extension

        x = np.asarray(x)
        return UnevenFiniteDifferenceWeights(m, z, x).T

    @classmethod
    def finite_difference_data(cls, grid, order, stencil, end_point_precision):
        """Constructs a finite-difference function that computes the nth derivative with a given width

                :param deriv:
                :type deriv:
                :param accuracy:
                :type accuracy:
                :return:
                :rtype:
                """

        outer_stencil = stencil + end_point_precision

        # we're gonna do just the 1D case for now
        left_pad = math.floor(stencil / 2)
        right_pad = math.ceil(stencil / 2)
        slices_outer = cls.get_grid_slices(grid, outer_stencil)
        slices_core = cls.get_grid_slices(grid, stencil)
        slices_left = slices_outer[:left_pad]
        slices_right = slices_outer[-right_pad:]
        coeffs = [
            [cls.get_weights(order, x[0], x) for x in slices_left],
            [cls.get_weights(order, x[left_pad], x) for x in slices_core],
            [cls.get_weights(order, x[-1], x) for x in slices_right]
            ]
        widths = [(outer_stencil, 0), (left_pad, right_pad), (0, outer_stencil)]

        return FiniteDifferenceData(coeffs, widths, order)

##########################################################################################
#
#                                    FiniteDifferenceData
#
##########################################################################################
class FiniteDifferenceData:
    """
    Holds the data used by to construct a finite difference matrix
    """
    def __init__(self, weights, widths, order):
        self._weights = self._clean_coeffs(weights)  # used to apply the weights to function values
        # NOTE: the coefficients should be supplied without 'h' if an evenly spaced grid
        # These will be forced back in later
        self._widths = self._clean_widths(widths)  # used to determine how to apply the coefficients
        self._order = self._clean_order(order)  # the order of the derivative

    # region Properties
    @property
    def weights(self):
        return tuple(tuple(x) for x in self._weights)
    @property
    def widths(self):
        return tuple(tuple(x) for x in self._widths)
    @property
    def order(self):
        return self._order
    #endregion

    @classmethod
    def _clean_coeffs(cls, cfs):
        try:
            cl, cc, cr = cfs
            islcr = True
        except ValueError:
            islcr = False
        if not islcr:
            raise TypeError("{}: coefficients {} is expected to be list of list coefficients on the left, center, and right",
                            cls.__name__, cfs)
        return cfs
    @classmethod
    def _clean_order(cls, order):
        if not isinstance(order, (int, np.integer)):
            raise TypeError("{}: order {} is expected to be an int",
                            cls.__name__, order)
        return order
    @classmethod
    def _clean_widths(cls, ws):

        if isinstance(ws, (int, float, np.integer, np.floating)):
            ws = ( (int(ws), int(ws)) )
        else:
            w2s = [ None ] * len(ws)
            for i, w in enumerate(ws):
                if isinstance(w, (int, float, np.integer, np.floating)):
                    w2s[i] = (int(w), int(w))
                elif w is None:
                    pass
                elif isinstance(w[0], (int, float, np.integer, np.floating)):
                    w2s[i] = (int(w[0]), int(w[1]))
                elif w is not None:
                    w2s[i] = cls._clean_widths(w)
            ws = tuple(w2s)
        return ws

##########################################################################################
#
#                                    FiniteDifferenceMatrix
#
##########################################################################################
class FiniteDifferenceMatrix:
    """
    Defines a matrix that can be applied to a regular grid of values to take a finite difference
    """

    def __init__(self,
                 finite_difference_data,
                 npts = None, mesh_spacing = None,
                 only_core = False, only_center = False,
                 mode = "sparse", dtype = "float64"
                 ):
        """
        :param finite_difference_data:
        :type finite_difference_data: FiniteDifferenceData
        :param npts:
        :type npts:
        :param mesh_spacing:
        :type mesh_spacing:
        :param only_core:
        :type only_core:
        :param only_center:
        :type only_center:
        :param mode:
        :type mode:
        """
        self.data = finite_difference_data
        self._npts = npts
        self._mesh_spacing = mesh_spacing
        self._only_core = only_core
        self._only_center = only_center
        self._mode = mode
        self._mat = None
        self._dtype = dtype

    #region Properties
    @property
    def weights(self):
        return self.data.weights
    @property
    def order(self):
        return self.data.order
    @property
    def npts(self):
        return self._npts
    @npts.setter
    def npts(self, val):
        if val != self._npts:
            self._mat = None
            self._npts = val
    @property
    def mesh_spacing(self):
        return self._mesh_spacing
    @mesh_spacing.setter
    def mesh_spacing(self, val):
        if val != self._mesh_spacing:
            self._mat = None
            self._mesh_spacing = val
    @property
    def only_core(self):
        return self._only_core
    @only_core.setter
    def only_core(self, val):
        if val != self._only_core:
            self._mat = None
            self._only_core = val
    @property
    def only_center(self):
        return self._only_center
    @only_center.setter
    def only_center(self, val):
        if val != self._only_center:
            self._mat = None
            self._only_center = val
    @property
    def mode(self):
        return self._mode
    @mode.setter
    def mode(self, val):
        if val != self._mode:
            self._mat = None
            self._mode = val
    @property
    def dtype(self):
        return self._dtype
    @dtype.setter
    def dtype(self, val):
        if val != self._dtype:
            self._mat = None
            self._dtype = val
    @property
    def matrix(self):
        if self._mat is None:
            self._mat = self.fd_matrix()
        return self._mat
    #endregion

    def fd_matrix(self):
        """Builds a 1D finite difference matrix for a set of boundary weights, central weights, and num of points
        Will look like:
            b1 b2 b3 ...
            w1 w2 w3 ...
            0  w1 w2 w3 ...
            0  0  w1 w2 w3 ...
                 ...
                 ...
                 ...
                    .... b3 b2 b1
        :return: fd_mat
        :rtype: np.ndarray | sp.csr_matrix
        """
        npts = int(self.npts)
        only_core = self.only_core
        only_center = self.only_center
        o = self.order
        h = self.mesh_spacing
        dtype = self.dtype
        if h is not None:
            c_left, c_center, c_right = [np.array(x, dtype=dtype) / (h**o) for x in self.weights]
        else:
            c_left, c_center, c_right = [np.array(x, dtype=dtype) for x in self.weights]
        mode = self.mode

        if isinstance(c_center[0], (int, float, np.integer, np.floating)):
            fdm = self._fdm_regular(c_left, c_center, c_right, npts, only_core, only_center, mode, dtype)
        else:
            fdm = self._fdm_irregular(c_left, c_center, c_right, npts, only_core, only_center, mode, dtype)

        return fdm

    @classmethod
    def _fdm_irregular(cls, c_left, c_center, c_right, npts, only_core, only_center, mode, dtype):
        lcc = len(c_center)
        if only_center:
            shape = (1, npts)
            if mode == "sparse":
                fdm = sparse.lil_matrix(shape)
            else:
                fdm = np.zeros(shape)
            lcf = len(c_left)
            mid = math.floor(lcc)
            x = c_center[mid]
            p = lcf+mid
            fdm[p:p+len(x)] = x
        elif only_core:
            lcf = len(c_left)
            shape = (npts - lcc + (lcc % 2), npts)
            if mode == "sparse":
                fdm = sparse.lil_matrix(shape)
            else:
                fdm = np.zeros(shape)
            for i, x in enumerate(c_center):
                p = lcf+i
                fdm[i, p:p+lcc] = x
        else:
            shape = (npts, npts)
            if mode == "sparse":
                fdm = sparse.lil_matrix(shape)
            else:
                fdm = np.zeros(shape)
            lcf = len(c_left)
            for i, x in enumerate(c_left):
                fdm[i, i:i+lcf] = x
            for j, x in enumerate(c_center):
                fdm[j, j:j+lcc] = x
            lcr = len(c_right)
            for k, x in enumerate(c_right):
                fdm[-k, -(k+lcr):-k] = x

        return fdm

    @classmethod
    def _fdm_regular(cls, c_left, c_center, c_right, npts, only_core, only_center, mode, dtype):
        lcc = len(c_center)
        if only_center:

            if mode == "sparse":
                fdm = sparse.lil_matrix((1, npts), dtype=dtype)
            else:
                fdm = np.zeros((1, npts))
            grid_mid = math.floor((npts - lcc + (lcc % 2))/2)
            x = c_center
            p = grid_mid
            fdm[0, p:p + len(x)] = x
        elif only_core:
            fdm = cls._fdm_core(c_center, npts - lcc + (lcc % 2), npts, mode=mode)
        else:
            bound_l = min(int(math.floor(lcc / 2)), int(math.floor(npts / 2)))
            # TODO: sometimes the grid-point stencils overrun the grid I actually have (for small grids)
            # which means I need to handle these boundary cases better

            bdm_l = cls._fdm_core(c_left, bound_l, npts, mode=mode)
            bdm_r = cls._fdm_core(c_right, bound_l, npts, side="r", mode=mode)
            core = cls._fdm_core(c_center, npts - lcc + (lcc % 2), npts, mode=mode)
            if mode == "sparse":
                fdm = sparse.vstack((bdm_l, core, bdm_r))
                # print("\n", bdm_r.todense())
            else:
                fdm = np.array(np.vstack((bdm_l, core, bdm_r)))
                # print("\n", fdm)

        return fdm

    @classmethod
    def _fdm_core(cls, a, n1, n2, side="l", mode = "dense"):
        # pulled from https://stackoverflow.com/a/52464135/5720002
        # returns the inner blocks of the FD mat
        if mode == "sparse":
            offs = (n2-len(a)-n1+1) + np.arange(len(a)) if side == "r" else np.arange(len(a))
            mmm = sparse.diags(a, offsets=offs, shape=(n1, n2), dtype=a.dtype)
        else:
            a = np.asarray(a)
            p = np.zeros(n2, dtype=a.dtype)
            b = np.concatenate((p, a, p))
            s = b.strides[0]
            strided = np.lib.stride_tricks.as_strided
            if side == "r":
                mmm = strided(b[len(a) + n1 - 1:], shape=(n1, n2), strides=(-s, s))
            else:
                mmm = strided(b[n2:], shape=(n1, n2), strides=(-s, s))
        return mmm

##########################################################################################
#
#                                    finite_difference
#
##########################################################################################

def finite_difference(grid, values, order,
                      accuracy=2,
                      stencil=None,
                      end_point_accuracy=1,
                      axes=None,
                      only_core=False,
                      only_center=False,
                      dtype="float64",
                      **kw
                      ):
    """Computes a finite difference derivative for the values on the grid

    :param grid: the grid of points for which the vlaues lie on
    :type grid: np.ndarray
    :param values: the values on the grid
    :type values: np.ndarray
    :param order: order of the derivative to compute
    :type order: int | Iterable[int]
    :param stencil: number of points to use in the stencil
    :type stencil: int | Iterable[int]
    :param accuracy: approximate accuracy of the derivative to request (overridden by `stencil`)
    :type accuracy: int | Iterable[int]
    :param end_point_accuracy: extra stencil points to use on the edges
    :type end_point_accuracy: int | Iterable[int]
    :param end_point_accuracy: extra stencil points to use on the edges
    :param axes: which axes to perform the successive derivatives over (defaults to the first _n_ axes)
    :type axes: int | Iterable[int]
    :param only_center: whether or not to only take the central value
    :type only_center: bool
    :param only_core: whether or not to avoid edge values where a different stencil would be used
    :type only_core: bool
    :return:
    :rtype:
    """
    gv = np.asarray(values)
    if 'end_point_precision' in kw:
        end_point_accuracy = kw['end_point_precision']
    if 'axis' in kw:
        axes = kw['axis']
    func = FiniteDifferenceFunction.from_grid(
        grid, order,
        accuracy = accuracy,
        stencil = stencil,
        end_point_accuracy = end_point_accuracy,
        axes=axes,
        dtype=dtype,
        only_core=only_core,
        only_center=only_center,
        **kw
    )
    return func(gv)
    # if func.gridtype != FiniteDifferenceFunction.IRREGULAR_GRID:
    #     return fdf(gv)
    # else:
    #     return fdf(gv)


import uuid

from .CoordinateSystem import CoordinateSystem
from .CoordinateSystemConverter import CoordinateSystemConverter
from ...Numputils import apply_pointwise
import weakref

__all__ = [
    "CompositeCoordinateSystem",
    "CompositeCoordinateSystemConverter"
]

#TODO: these should all be metaclasses but :shrag:
class CompositeCoordinateSystem(CoordinateSystem):
    """
    Defines a coordinate system that comes from applying a transformation
    to another coordinate system
    """

    _register_cache = weakref.WeakValueDictionary()
    def __init__(self, base_system, conversion, inverse_conversion=None, name=None, batched=None, pointwise=True, **opts):
        self.base_system = base_system
        self.conversion = conversion
        self.inverse_conversion = inverse_conversion
        self.pointwise = pointwise
        self.batched = batched if batched is not None else not pointwise
        super().__init__(**opts)
        self.name = self.canonical_name(name, conversion)
    @classmethod
    def canonical_name(cls, name, conversion):
        if name is None:
            if hasattr(conversion, 'name'):
                name = conversion.name
            else:
                name = str(uuid.uuid4()).replace("-", "")
        return name
        #
        #     return type('CompositeCoordinateSystem' + name, (cls,),
        #                 {'base_system': base_system, 'conversion': conversion, 'inverse': inverse})
        # # if self.base_system is None:
        # #     raise ValueError('{name} is a factory class and {name}.{method} should be used to register coordinate systems'.format(
        # #         name=type(self).__name__,
        # #         method='register'
        # #     ))
        # # super().__init__()

    @classmethod
    def register(cls, base_system, conversion, inverse_conversion=None, name=None, batched=None, pointwise=True, **opts):
        if (base_system, conversion) not in cls._register_cache:
            system_class = cls(base_system, conversion, inverse_conversion=inverse_conversion, name=name,
                               batched=batched, pointwise=pointwise, **opts)
            CompositeCoordinateSystemConverter(system_class).register()
            if system_class.inverse_conversion is not None:
                CompositeCoordinateSystemConverter(system_class, direction='inverse').register()
            cls._register_cache[(base_system, conversion)] = system_class
        return cls._register_cache[(base_system, conversion)]
    def unregister(self):
        raise NotImplementedError("destructor not here yet")
    def __repr__(self):
        return "{}({}, {})".format(
            type(self).__name__,
            self.base_system,
            self.name
        )


class CompositeCoordinateSystemConverter(CoordinateSystemConverter):
    def __init__(self, system, direction='forward'):
        self.system = system
        self.direction = direction
    @property
    def types(self):
        if self.direction == 'inverse':
            return (self.system, self.system.base_system)
        else:
            return (self.system.base_system, self.system)
    def get_conversion(self):
        if self.direction == 'forward':
            convertser = self.system.conversion
        elif self.direction == 'inverse':
            convertser = self.system.inverse_conversion
        else:
            raise NotImplementedError("bad value for '{}': {}".format('direction', self.direction))
        return convertser
    def convert(self, coords, **kw):
        if self.system.pointwise:
            return apply_pointwise(self.get_conversion(), coords, **kw)
        else:
            return self.get_conversion()(coords, **kw)
    def convert_many(self, coords, **kw):
        if self.system.pointwise:
            return apply_pointwise(self.get_conversion(), coords, **kw)
        elif self.system.batched:
            return self.convert(coords, **kw)
        else:
            return super().convert_many(coords, **kw)





# OLD AND DEPRECATED
class JHTMLShadowDOMElement:
    """
    Provides a shadow DOM tree that can makes it easier to update
    chunks of a tree constructed from JupyterHTMLWrappers
    """
    def __init__(self, widg, wrapper=None, parent=None):
        self.widg = widg
        self.wrapper = wrapper
        self.parent = parent.dom if hasattr(parent, 'dom') else parent
        self._tree_cache = None

    def link(self, dom_el, recursive=False):
        dom_el.on_update = self.Refresher(self)
        dom_el.shadow = self
        if recursive:
            for c in dom_el:
                if isinstance(c, HTML.XMLElement):
                    self.link(c, recursive=True)
    @property
    def tree(self):
        return self.to_tree()
    def to_tree(self, refresh=False):
        if refresh:
            self._invalidate_cache()
        if self._tree_cache is None:
            if hasattr(self.widg, 'children'):
                children = []
                for c in self.widg.children:
                    if hasattr(c, 'dom'):
                        el = c.dom.to_tree()
                        if el is not None:
                            children.append(el)
                    elif hasattr(c, 'value'):
                        shadow = type(self)(c, wrapper=None, parent=self)
                        # self.link(elem)
                        # elem.shadow = self
                        # elem.on_update = self.Refresher(self, which=len(children))
                        el = shadow.to_tree()
                        if el is not None:
                            children.append(el)
                attrs = {}
                if len(self.widg._dom_classes) > 0:
                    attrs['cls'] = self.widg._dom_classes
                dom_el = HTML.Div(children, **attrs)
            elif hasattr(self.widg, 'value'):
                dom_el = HTML.parse(self.widg.value, strict=False)
                self.link(dom_el, recursive=True)
            else:
                dom_el = None
            self._tree_cache = dom_el
            if dom_el is not None:
                self.link(dom_el, recursive=False)
                # dom_el.shadow = self
                # dom_el.on_update = self.Refresher(self)
        return self._tree_cache
    def _invalidate_cache(self):
        self._tree_cache = None
        if self.parent is not None:
            self.parent._invalidate_cache()

    class Refresher:
        def __init__(self, dom_el, which=None):
            self.dom_el = dom_el
            self.which = which
        def __call__(self, el, key, value):
            if self.which is not None:
                self.dom_el.refresh_child(el, key, value)
            else:
                self.dom_el.refresh(el, key, value)
    def refresh(self, el, key, value):
        if hasattr(self.widg, 'children'):
            if isinstance(key, str):
                if key=='elements':
                    raise NotImplementedError("haven't handle full elem updates yet")
                elif key == 'attributes':
                    k = list(value.keys())
                    if 'class' in value:
                        classes = value['class']
                        if isinstance(classes, str) or hasattr(classes, 'tostring'):
                            classes = [classes]
                        self.widg._dom_classes = [str(c) for c in classes]
                        k.remove('class')
                    if len(k) > 0:
                        raise NotImplementedError("haven't handled attr updates to keys {} yet".format(k))
                elif key == 'class':
                    if isinstance(value, str) or hasattr(value, 'tostring'):
                        value = [value]
                    self.widg._dom_classes = [str(c) for c in value]
                else:
                    raise NotImplementedError("can't update attribute {}")
            else:
                # print(self.widg.children[key])
                self.widg.children[key].value = value.tostring()
                self.widg.children[key].dom.invalidate_cache()
                # print(self.widg.children[key])
        else:
            self.widg.value = el.tostring()
        self._invalidate_cache()
    def refresh_child(self, el, key, value, which):
        raise NotImplementedError("ugh")
        subwidg = self.widg.children[which]
        if hasattr(subwidg, 'children'):
            raise NotImplementedError("ugh")
        else:
            subwidg.value = el.tostring()
        self.invalidate_cache()

    def get_parent(self, n):
        for i in range(n):
            self = self.parent
        return self

    def find(self, path, find_element=True):
        return self.tree.find(path, find_element=find_element)
    def findall(self, path, find_element=True):
        return self.tree.findall(path, find_element=find_element)
    def iterfind(self, path, find_element=True):
        return self.tree.iterfind(path, find_element=find_element)
    def find_by_id(self, id, mode='first', parent=None, find_element=True):
        return self.tree.find_by_id(id, mode=mode, parent=parent, find_element=find_element)
    def __getitem__(self, item):
        return self.to_tree()[item]
class JupyterHTMLWrapper:
    """
    Provides simple access to Jupyter display utilities
    """

    reset_attrs = [
        'background',
        'color',
        'font',
        'font-family'
    ]
    @classmethod
    def load_styles(cls):
        """
        Embeds widget restyle definitions into the active notebook

        :return:
        :rtype:
        """
        from IPython.core.display import HTML as IPyHTML
        custom_styles = [
            CSS.construct(
                '.jhtml-reset .jhtml-html',
                '.jhtml-reset .jhtml-parent',
                display='inline-block'
            )
        ]
        return IPyHTML(HTML.Style("\n".join(s.tostring() for s in custom_styles)).tostring())
    stripped_classes = ['lm-Widget', 'p-Widget', 'lm-Panel', 'p-Panel', 'jupyter-widgets', 'widget-container', 'widget-box', 'widget-html']
    protecting_classes = ['widget-slider']
    stripped_unprotected = ['widget-inline-hbox']
    @classmethod
    def get_class_stripper_js(cls):
        import uuid
        this = "stripper"+str(uuid.uuid1()).replace("-", "")[:8]
        script = """
        var {id} = document.getElementById('{id}');
        var {id}siblings = {id}.parentNode.parentNode.parentNode.childNodes;
        function {id}strip(node) {{
           if (typeof node.classList != 'undefined') {{
               node.classList.remove({stripped_classes});
               // ipython
               if (!{protecting_classes}.some(e=>node.classList.contains(e))) {{
                 node.classList.remove({stripped_unprotected});
               }}
           }}
           if (typeof node.classList != 'undefined') {{
               for (const child of node.childNodes) {{
                 {id}strip(child);
               }}
           }}
           // if (node.tagName === 'DIV') {{ 
           //    // something about chained conditions is giving as syntax error...
           //    if (node.classList.length == 1) {{
           //       if (node.classList[0] == 'jhtml-reset') {{
           //          node.replaceWith(...node.childNodes);
           //       }}
           //    }}
           // }}
        }}
        for (const sibling of {id}siblings) {{
            {id}strip(sibling);
        }}
        {id}.parentNode.parentNode.remove();
                    """.format(id=this,
                               stripped_classes=", ".join("'{}'".format(s) for s in cls.stripped_classes),
                               protecting_classes="[" +", ".join("'{}'".format(s) for s in cls.protecting_classes)+"]",
                               stripped_unprotected=", ".join("'{}'".format(s) for s in cls.stripped_unprotected)
                               )
        return script, this
    @classmethod
    def get_class_stripper(cls):
        stripper, this = cls.get_class_stripper_js()
        return JupyterAPIs.get_display_api().HTML('<script id="{this}">{script}</script>'.format(script=stripper, this=this))

    cls = None #for easy overloads later
    tag = None
    layout = None
    container = False # overloaded later
    def __init__(self, *elements,
                 tag=None,
                 event_handlers=None,
                 layout=None,
                 extra_classes=None,
                 cls=None,
                 debug_pane=None,
                 **styles
                 ):
        if len(elements) == 1 and isinstance(elements[0], (list, tuple)):
            elements = elements[0]
        self.elements = elements
        self.event_handlers = event_handlers
        if event_handlers is not None and not hasattr(self.event_handlers, 'items'):
            self.event_handlers = {
                'click':self.event_handlers,
                'mouseenter':self.event_handlers,
                'mouseleave':self.event_handlers,
                'keydown':self.event_handlers,
                'keyup':self.event_handlers
            }
        self.tag = tag if tag is not None else self.tag
        self.layout = self.layout if layout is None else layout
        self.styles = styles
        if extra_classes is None:
            extra_classes = self.cls
        if cls is not None:
            if hasattr(cls, 'value'):
                cls = cls.value
            if isinstance(cls, str):
                cls = cls.split()

            if extra_classes is None:
                extra_classes = cls
            else:
                if hasattr(extra_classes, 'value'):
                    extra_classes = extra_classes.value
                if isinstance(extra_classes, str):
                    extra_classes = extra_classes.split()
                extra_classes = list(extra_classes) + list(cls)
        if hasattr(extra_classes, 'value'):
            extra_classes = extra_classes.value
        if isinstance(extra_classes, str):
            extra_classes = extra_classes.split()
        if extra_classes is not None:
            extra_classes = [e.value if hasattr(e, 'value') else e for e in extra_classes]
        self.extra_classes = extra_classes

        self._parents = weakref.WeakSet()
        self._widget_cache = None
        if debug_pane is None:
            debug_pane = DefaultOutputArea.get_default()
        self.debug_pane = debug_pane

    @staticmethod
    def _handle_event(e, event_handlers, self, widg):
        try:
            handler = event_handlers[e['type']]
        except KeyError:
            pass
        else:
            handler(e, self, widg)
    def _event_handler(self, widg):
        def handler(e, output=self.debug_pane):
            with output:
                return self._handle_event(e, self.event_handlers, self, widg)
        return handler


    inherited_classes = ['d-inline', 'd-flex', 'd-inline-block', 'd-block', 'd-none'] # explicitly Bootstrap oriented for now...
    inherited_class_tag_map = { # this is _unlikely_ to be overwritten...
        'span':['d-inline-block'],
        None:['d-inline-block']
    }
    inherited_class_cls_map = { # again _unlikely_ to be overwritten
        'btn': ['d-inline-block']
    }
    def _convert(self, x, parent=None):
        widget_classes = ['jhtml-reset', 'jhtml-html']
        if isinstance(x, (HTML.XMLElement, HTML.ElementModifier)):
            # to get widget formatting right for container classes...
            try:
                mapped_widget = self.base_map[type(x)]
            except KeyError:
                mapped_widget = None
                needs_mods = False
            else:
                try:
                    needs_mods = mapped_widget.container
                except AttributeError:
                    needs_mods = False
                if mapped_widget.cls is None:
                    mapped_widget = type(x)
                if needs_mods:
                    x = x.remove_class(mapped_widget.cls)

            # now we propagate up inherited classes
            try:
                cls = x.attrs['class']
            except (AttributeError, KeyError):
                cls = None
            if cls is None:
                cls = []
            elif isinstance(cls, str):
                cls = cls.split()
            cls = list(cls)

            tag = x.tag.lower()
            if tag in self.inherited_class_tag_map:
                cls = list(self.inherited_class_tag_map[tag]) + cls
            for c in cls:
                if c in self.inherited_class_cls_map:
                    cls = list(self.inherited_class_cls_map[c]) + cls

            extras = [e for e in cls if e in self.inherited_classes]
            if len(extras) > 0:
                for e in extras:
                    widget_classes.append(e)
                x = x.remove_class(*extras)

            x = x.tostring()
        else:
            if isinstance(x, str):
                cls = list(self.inherited_class_tag_map[None])
                extras = [e for e in cls if e in self.inherited_classes]
                if len(extras) > 0:
                    for e in extras:
                        widget_classes.append(e)
            mapped_widget = None
            needs_mods = False

        if isinstance(x, str):
            w = JupyterAPIs().widgets_api.HTML(x)
            if needs_mods:
                cls = mapped_widget.cls
                if cls is None:
                    cls = []
                elif hasattr(cls, 'tostring'):
                    cls = cls.tostring
                if isinstance(cls, str):
                    cls = cls.split()
                extra_classes = list(cls)
                for c in extra_classes:
                    if parent.extra_classes is None or c not in parent.extra_classes:
                        # print(c, parent.extra_classes)
                        w.add_class(str(c))
            for c in widget_classes:
                w.add_class(c)
        elif isinstance(x, JupyterHTMLWrapper):
            w = x.to_widget(parent=self)
        else:
            w = x
        return w

    @classmethod
    def manage_styles(cls, styles, validate=True):
        layout_props = {
            'height',
            'width',
            'max_height',
            'max_width',
            'min_height',
            'min_width',
            'visibility',
            'display',
            'overflow',
            'model',
            'border',
            'margin',
            'padding',
            'top',
            'left',
            'bottom',
            'right',
            'order',
            'flex_flow',
            'align_items',
            'flex',
            'align_self',
            'align_content',
            'justify_content',
            'grid_auto_columns',
            'grid_auto_flow',
            'grid_auto_rows',
            'grid_gap',
            'grid_template',
            'grid_row',
            'grid_column'
        }
        style_props = {
            "color",
            "background",
            'font_family',
            'font_size',
            'font_style',
            'font_variant',
            'font_weight',
            'text_color',
            'text_decoration'
        }

        if styles is None:
            styles = {}

        layout_map = {}
        style_map = {}
        for k in styles:
            if k in layout_props:
                layout_map[k] = styles[k]
            elif k in style_props:
                style_map[k] = styles[k]
            elif validate:
                raise ValueError("as designed JupyterLab doesn't support the {} attribute".format(
                    k
                ))
        return {'style':style_map, 'layout':layout_map}

    def _invalidate_cache(self):
        if self._widget_cache is not None:
            self._widget_cache = None
            for p in tuple(self._parents):
                p.invalidate_cache()
                self._parents.remove(p)
    def to_widget(self, parent=None):
        if parent is not None:
            self._parents.add(parent)
        if self._widget_cache is None:
            layout = self.layout
            if isinstance(self.elements[0], (list, tuple)):
                widgets = [
                    [ self._convert(x, parent=self) for x in y ]
                    for y in self.elements
                ]
            else:
                widgets = [self._convert(x, parent=self) for x in self.elements]

            if 'style' in self.styles:
                props = self.manage_styles(self.styles['style'])
            elif len(self.styles) > 0:
                raise ValueError("JupyterLab doesn't support {} attributes on widgets".format(list(self.styles.keys())))
            else:
                props = {}

            if layout is None:
                layout = 'box'
            if layout == 'grid':
                widg = JupyterAPIs().widgets_api.GridBox(widgets, **props)
            elif layout == 'row':
                widg = JupyterAPIs().widgets_api.HBox(widgets, **props)
            elif layout == 'column':
                widg = JupyterAPIs().widgets_api.VBox(widgets, **props)
            else:
                def flatten(x, base):
                    if not isinstance(x, JupyterAPIs().widgets_api.Widget):
                        for y in x:
                            flatten(y, base)
                    else:
                        base.append(x)
                new = []; flatten(widgets, new)
                widg = JupyterAPIs().widgets_api.Box(new, **props)

            from IPython.core.display import display as IPyDisplay
            widg.add_class('jhtml-reset')
            widg.add_class('jhtml-wrapper')
            if self.extra_classes is not None:
                for c in self.extra_classes:
                    widg.add_class(c)
            if self.event_handlers is not None:
                listener = JupyterAPIs().events_api.Event(source=widg, watched_events=list(self.event_handlers.keys()))
                listener.on_dom_event(self._event_handler(widg))
                widg.listener = listener
                widg.add_class('jhtml-event-handler')
                # if self.extra_classes is None or len(self.extra_classes) == 0:
                for child in widg.children:
                    if not hasattr(child, 'listener'):
                        listener = JupyterAPIs().events_api.Event(source=child,
                                                                  watched_events=list(self.event_handlers.keys()))
                        listener.on_dom_event(self._event_handler(child))
                        child.listener = listener
                        child.add_class('jhtml-event-handler')
            html_widget = JupyterAPIs.get_widgets_api().HTML
            if (
                    len(widg.children)==1
                    and isinstance(widg.children[0], html_widget)
            ):
                widg.add_class('jhtml-parent')
                for c in widg.children[0]._dom_classes:
                    if c in self.inherited_classes:
                        widg.add_class(c)

            # for discoverability in callbacks
            widg.dom = JHTMLShadowDOMElement(widg, wrapper=self)
            if hasattr(widg, 'children'):
                for x in widg.children:
                    if not hasattr(x, 'dom'):
                        x.dom = JHTMLShadowDOMElement(x, wrapper=None, parent=widg.dom)
                    else:
                        x.dom.parent = widg.dom
            widg.display = lambda w=widg:self.display_widget(w)
            self._widget_cache = widg
        return self._widget_cache
    @classmethod
    def display_widget(cls, w):
        return JupyterAPIs.get_display_api().display(w, cls.get_class_stripper())
    def display(self):
        return JupyterAPIs.get_display_api().display(self.to_widget(), self.get_class_stripper())
    def _ipython_display_(self):
        return self.display()

    def find(self, path, find_element=True):
        return self.to_widget().dom.find(path, find_element=find_element)
    def findall(self, path, find_element=True):
        return self.to_widget().dom.findall(path, find_element=find_element)
    def iterfind(self, path, find_element=True):
        return self.to_widget().dom.iterfind(path, find_element=find_element)
    def find_by_id(self, id, mode='first', parent=None, find_element=True):
        return self.to_widget().dom.find_by_id(id, mode=mode, parent=parent, find_element=find_element)

    _widget_sources = []
    _base_map = None
    @classmethod
    def load_base_map(cls):
        if cls._base_map is None:
            cls._base_map = {}
            for src in cls._widget_sources:
                for v in src.__dict__.values():
                    if hasattr(v, 'base'):
                        cls._base_map[v.base] = v
        return cls._base_map
    @property
    def base_map(self):
        return self.load_base_map()
class JupyterHTMLWidgets:
    """
    Provides convenience constructors for HTML components
    """
    @classmethod
    def load(cls):
        return JupyterHTMLWrapper.load_styles()

    class WrappedElement(JupyterHTMLWrapper):
        base = None
        container = False
        def __init__(self, *elems, base=None, event_handlers=None, container=None, layout=None, extra_classes=None, **attrs):
            if base is None:
                base = self.base
            self.base = base
            self.container = self.container if container is None else container
            Widget = JupyterAPIs.get_widgets_api().Widget
            if any(isinstance(e, (JupyterHTMLWrapper, Widget)) for e in elems):
                elem = elems
                if hasattr(base, 'cls'):
                    cls = base.cls
                    if cls is None:
                        cls = []
                    elif isinstance(cls, str):
                        cls = cls.split()
                    extras = list(cls)

                    if extra_classes is None:
                        extra_classes = []
                    elif isinstance(extra_classes, str):
                        extra_classes = extra_classes.split()
                    extra_classes = list(extra_classes) + extras

                if not hasattr(base, 'tag') or base.tag != "div":
                    raise ValueError("{} can't be a proper widget base as JupyterLab is designed (i.e. it can't have widgets as children)".format(base))
            else:
                elem = base(*elems, **attrs)
                if self.container and hasattr(base, 'cls'):
                    cls = base.cls
                    if cls is None:
                        cls = []
                    elif isinstance(cls, str):
                        cls = cls.split()
                    extras = list(cls)

                    if extra_classes is None:
                        extra_classes = []
                    elif isinstance(extra_classes, str):
                        extra_classes = extra_classes.split()
                    extra_classes = list(extra_classes) + extras
                    elem = elem.remove_class(*extras)
                attrs = {}
            super().__init__(elem, event_handlers=event_handlers, layout=layout, extra_classes=extra_classes, **attrs)
        def copy(self):
            import copy
            new = copy.copy(self)
            new.elements = [(x.copy() if hasattr(x, 'copy') else x) for x in new.elements]
            new._widget_cache = None
            new._parents = weakref.WeakSet()
            return new
        def add_child_class(self, *cls, copy=True):
            if copy:
                new = self.copy()
                return new.add_child_class(*cls, copy=False)
            else:
                self.elements = [x.add_class(*cls) for x in self.elements]
                return self
        def add_class(self, *cls, copy=True):
            if copy:
                new = self.copy()
                return new.add_class(*cls, copy=False)
            else:
                if self.extra_classes is None:
                    self.extra_classes = [str(x) for x in cls]
                else:
                    self.extra_classes = [str(x) for x in self.extra_classes]
                for c in cls:
                    if c not in self.extra_classes:
                        self.extra_classes.append(str(c))
                self._invalidate_cache()
                return self
        def remove_class(self, *cls, copy=True):
            if copy:
                new = self.copy()
                return new.remove_class(*cls, copy=False)
            else:
                if self.extra_classes is not None:
                    self.extra_classes = [str(x) for x in self.extra_classes]
                for c in cls:
                    try:
                        self.extra_classes.remove(str(c))
                    except ValueError:
                        pass
                self._invalidate_cache()
                return self
        def add_styles(self, copy=True, **sty):
            if copy:
                new = self.copy()
                return new.add_styles(copy=False, **sty)
            else:
                self.elements = [x.add_styles(**sty) for x in self.elements]
                self._invalidate_cache()
                return self
    class ContainerWrapper(WrappedElement): container = True
    class Abbr(WrappedElement): base=HTML.Abbr
    class Address(WrappedElement): base=HTML.Address
    class Anchor(WrappedElement): base=HTML.Anchor
    A = Anchor
    class Area(WrappedElement): base=HTML.Area
    class Article(WrappedElement): base=HTML.Article
    class Aside(WrappedElement): base=HTML.Aside
    class Audio(WrappedElement): base=HTML.Audio
    class B(WrappedElement): base=HTML.B
    class Base(WrappedElement): base=HTML.Base
    class BaseList(WrappedElement): base=HTML.BaseList
    class Bdi(WrappedElement): base=HTML.Bdi
    class Bdo(WrappedElement): base=HTML.Bdo
    class Blockquote(WrappedElement): base=HTML.Blockquote
    class Body(WrappedElement): base=HTML.Body
    class Bold(WrappedElement): base=HTML.Bold
    class Br(WrappedElement): base=HTML.Br
    class Button(WrappedElement): base=HTML.Button
    class Canvas(WrappedElement): base=HTML.Canvas
    class Caption(WrappedElement): base=HTML.Caption
    class Cite(WrappedElement): base=HTML.Cite
    class ClassAdder(WrappedElement): base=HTML.ClassAdder
    class ClassRemover(WrappedElement): base=HTML.ClassRemover
    class Code(WrappedElement): base=HTML.Code
    class Col(WrappedElement): base=HTML.Col
    class Colgroup(WrappedElement): base=HTML.Colgroup
    class Data(WrappedElement): base=HTML.Data
    class Datalist(WrappedElement): base=HTML.Datalist
    class Dd(WrappedElement): base=HTML.Dd
    class Del(WrappedElement): base=HTML.Del
    class Details(WrappedElement): base=HTML.Details
    class Dfn(WrappedElement): base=HTML.Dfn
    class Dialog(WrappedElement): base=HTML.Dialog
    class Div(WrappedElement): base=HTML.Div
    class Dl(WrappedElement): base=HTML.Dl
    class Dt(WrappedElement): base=HTML.Dt
    class ElementModifier(WrappedElement): base=HTML.ElementModifier
    class Em(WrappedElement): base=HTML.Em
    class Embed(WrappedElement): base=HTML.Embed
    class Fieldset(WrappedElement): base=HTML.Fieldset
    class Figcaption(WrappedElement): base=HTML.Figcaption
    class Figure(WrappedElement): base=HTML.Figure
    class Footer(WrappedElement): base=HTML.Footer
    class Form(WrappedElement): base=HTML.Form
    class Head(WrappedElement): base=HTML.Head
    class Header(WrappedElement): base=HTML.Header
    class Heading(WrappedElement): base=HTML.Heading
    class Hr(WrappedElement): base=HTML.Hr
    class Iframe(WrappedElement): base=HTML.Iframe
    class Image(WrappedElement): base=HTML.Image
    class Img(WrappedElement): base=HTML.Img
    class Input(WrappedElement): base=HTML.Input
    class Ins(WrappedElement): base=HTML.Ins
    class Italic(WrappedElement): base=HTML.Italic
    class Kbd(WrappedElement): base=HTML.Kbd
    class Label(WrappedElement): base=HTML.Label
    class Legend(WrappedElement): base=HTML.Legend
    class Li(WrappedElement): base=HTML.Li
    class Link(WrappedElement): base=HTML.Link
    class List(WrappedElement): base=HTML.List
    class ListItem(WrappedElement): base=HTML.ListItem
    class Main(WrappedElement): base=HTML.Main
    class Map(WrappedElement): base=HTML.Map
    class Mark(WrappedElement): base=HTML.Mark
    class Meta(WrappedElement): base=HTML.Meta
    class Meter(WrappedElement): base=HTML.Meter
    class Nav(WrappedElement): base=HTML.Nav
    class Noscript(WrappedElement): base=HTML.Noscript
    class NumberedList(WrappedElement): base=HTML.NumberedList
    class Object(WrappedElement): base=HTML.Object
    class Ol(WrappedElement): base=HTML.Ol
    class Optgroup(WrappedElement): base=HTML.Optgroup
    class Option(WrappedElement): base=HTML.Option
    class Output(WrappedElement): base=HTML.Output
    class P(WrappedElement): base=HTML.P
    class Param(WrappedElement): base=HTML.Param
    class Picture(WrappedElement): base=HTML.Picture
    class Pre(WrappedElement): base=HTML.Pre
    class Progress(WrappedElement): base=HTML.Progress
    class Q(WrappedElement): base=HTML.Q
    class Rp(WrappedElement): base=HTML.Rp
    class Rt(WrappedElement): base=HTML.Rt
    class Ruby(WrappedElement): base=HTML.Ruby
    class S(WrappedElement): base=HTML.S
    class Samp(WrappedElement): base=HTML.Samp
    class Script(WrappedElement): base=HTML.Script
    class Section(WrappedElement): base=HTML.Section
    class Select(WrappedElement): base=HTML.Select
    class Small(WrappedElement): base=HTML.Small
    class Source(WrappedElement): base=HTML.Source
    class Span(WrappedElement): base=HTML.Span
    class Strong(WrappedElement): base=HTML.Strong
    class Style(WrappedElement): base=HTML.Style
    class StyleAdder(WrappedElement): base=HTML.StyleAdder
    class Sub(WrappedElement): base=HTML.Sub
    class SubHeading(WrappedElement): base=HTML.SubHeading
    class SubsubHeading(WrappedElement): base=HTML.SubsubHeading
    class SubsubsubHeading(WrappedElement): base=HTML.SubsubsubHeading
    class SubHeading5(WrappedElement): base=HTML.SubHeading5
    class SubHeading6(WrappedElement): base=HTML.SubHeading6
    class Summary(WrappedElement): base=HTML.Summary
    class Sup(WrappedElement): base=HTML.Sup
    class Svg(WrappedElement): base=HTML.Svg
    class Table(WrappedElement): base=HTML.Table
    class TableBody(WrappedElement): base=HTML.TableBody
    class TableHeading(WrappedElement): base=HTML.TableHeading
    class TableItem(WrappedElement): base=HTML.TableItem
    class TableRow(WrappedElement): base=HTML.TableRow
    class TagElement(WrappedElement): base=HTML.TagElement
    class Tbody(WrappedElement): base=HTML.Tbody
    class Td(WrappedElement): base=HTML.Td
    class Template(WrappedElement): base=HTML.Template
    class Text(WrappedElement): base=HTML.Text
    class Textarea(WrappedElement): base=HTML.Textarea
    class Tfoot(WrappedElement): base=HTML.Tfoot
    class Th(WrappedElement): base=HTML.Th
    class Thead(WrappedElement): base=HTML.Thead
    class Time(WrappedElement): base=HTML.Time
    class Title(WrappedElement): base=HTML.Title
    class Tr(WrappedElement): base=HTML.Tr
    class Track(WrappedElement): base=HTML.Track
    class U(WrappedElement): base=HTML.U
    class Ul(WrappedElement): base=HTML.Ul
    class Var(WrappedElement): base=HTML.Var
    class Video(WrappedElement): base=HTML.Video
    class Wbr(WrappedElement): base=HTML.Wbr

    class OutputArea(JupyterHTMLWrapper):
        def __init__(self, *elements, autoclear=False, event_handlers=None, layout=None, extra_classes=None, cls=None, **styles):
            if len(elements) == 1 and isinstance(elements, (list, tuple)):
                elements = elements[0]
            self.autoclear = autoclear
            self.output = JupyterAPIs.get_widgets_api().Output()
            elements = list(elements) + [self.output]
            super().__init__(*elements, tag='div', event_handlers=event_handlers, layout=layout, extra_classes=extra_classes, cls=cls, **styles)
        def print(self, *args, **kwargs):
            with self:
                print(*args, **kwargs)
        def display(self, *args):
            with self:
                JupyterAPIs().display_api.display(*args)
        def clear(self):
            self.output.clear_output()
        def __enter__(self):
            if self.autoclear:
                self.output.clear_output()
            return self.output.__enter__()
        def __exit__(self, exc_type, exc_val, exc_tb):
            return self.output.__exit__(exc_type, exc_val, exc_tb)
JupyterHTMLWrapper._widget_sources.append(JupyterHTMLWidgets)


"""
Represents an n-dimensional grid, used by Interpolator and (eventually) FiniteDifferenceFunction to automatically
know what kind of data is fed in
"""

import numpy as np, enum

__all__ = [ "Mesh", "MeshType" ]

# we define a bunch of MeshType attributes so that the string names can change up or whatever but the interface
# can remain consistent
class MeshType(enum.Enum):
    Regular = "regular" # direct product with consistent mesh spacings
    Structured = "structured" # direct product grids
    Unstructured = "unstructured" # totally unstructured data points
    SemiStructured = "semistructured" # almost full direct products
    Indeterminate = "indeterminate" # couldn't assess

class MeshError(Exception):
    ...

class Mesh(np.ndarray):
    """
    A general Mesh class representing data points in n-dimensions
    in either a structured, unstructured, or semi-structured manner.
    Exists mostly to provides a unified interface to difference FD and Surface methods.
    """

    # just to make it accessible through Mesh
    MeshError = MeshError
    MeshType = MeshType

    _allow_indeterminate = False
    # subclassing np.ndarray is weird... maybe I don't even want to do it... but I _think_ I do
    def __new__(cls,
                data,
                mesh_type=None,
                allow_indeterminate=None
                ):
        """
        :param griddata: the raw grid-point data the mesh uses
        :type griddata: np.ndarray
        :param mesh_type: the type of mesh we have
        :type mesh_type: None | str
        :param opts:
        :type opts:
        """
        data = np.asarray(data)
        if cls._is_meshgrid(data):
            # we turn mesh grids into grid point grids...
            data = np.moveaxis(data, 0, data.ndim-1)
        aid = cls._allow_indeterminate
        try:
            cls._allow_indeterminate = allow_indeterminate # gotta temp disable this...
            self = data.view(cls)
        finally:
            cls._allow_indeterminate = aid
        self.allow_indeterminate = allow_indeterminate
        self.mesh_type = self.get_mesh_type(self) if mesh_type is None else mesh_type
        self._spacings = None
        return self

    def __init__(self, *args, **kwargs):
        # empty init call because numpy is weird
        ...

    def __array_finalize__(self, mesh):
        # basically just a validator...
        if mesh is None:
            return None

        self.mesh_type = getattr(mesh, "mesh_type", None)
        if self.mesh_type is None:
            self.mesh_type = self.get_mesh_type(self)

        self.allow_indeterminate = getattr(mesh, "allow_indeterminate", None)
        aid = self.allow_indeterminate
        if aid is None:
            aid = self._allow_indeterminate
        if not aid and self.mesh_type is MeshType.Indeterminate:
            raise MeshError("indeterminate MeshType, but `allow_indeterminate` turned off")

    @property
    def mesh_spacings(self):
        if self._spacings is None:
            self._spacings = self.get_mesh_spacings(self)
        return self._spacings
    @property
    def subgrids(self):
        if (
                self.mesh_type == MeshType.Regular
                or self.mesh_type == MeshType.Structured
                # or self.mesh_type == MeshType.SemiStructured
        ):
            return self.get_mesh_subgrids(self)
        else:
            return None
    @property
    def bounding_box(self):
        gps = self.get_gridpoints(self).T
        return [(np.min(g), np.max(g)) for g in gps]
    @property
    def dimension(self):
        """Returns the dimension of the grid (not necessarily ndim)

        :return:
        :rtype: int
        """
        return self.shape[-1]
    @property
    def npoints(self):
        """Returns the number of gridpoints in the mesh

        :return:
        :rtype: int
        """
        return self.get_npoints(self)

    @property
    def gridpoints(self):
        """Returns the flattened set of gridpoints for a structured tensor grid and otherwise just returns the gridpoints

        :return:
        :rtype:
        """
        return self.get_gridpoints(self)

    @classmethod
    def get_npoints(cls, g):
        """Returns the number of gridpoints in the grid

        :param g:
        :type g: np.ndarray
        :return:
        :rtype: int
        """
        return np.prod(g.shape[:-1])
    @classmethod
    def get_gridpoints(cls, g):
        """Returns the gridpoints in the grid

        :param g:
        :type g: np.ndarray
        :return:
        :rtype: int
        """
        if isinstance(g, Mesh):
            g = g.view(np.ndarray)
        return np.reshape(g, (cls.get_npoints(g), g.shape[-1]))

    @classmethod
    def get_mesh_subgrids(cls, grid, tol=None):
        """
        Returns the subgrids for a mesh
        :param grid:
        :type grid:
        :param tol:
        :type tol:
        :return:
        :rtype:
        """

        if isinstance(grid, Mesh):
            grid = grid.view(np.ndarray)
        if tol is None:
            tol = grid.dtype.itemsize

        if grid.ndim == 1:
            _, idx = np.unique(np.round(grid, tol), return_index=True)
            return [grid[np.sort(idx),]]
        else:
            meshes = np.moveaxis(grid, len(grid.shape)-1, 0)
            def _pull_u(g):
                flat = g.flatten()
                _, idx = np.unique(g.flatten(), return_index=True)
                return flat[np.sort(idx),]
            return [_pull_u(np.round(m, tol)) for m in meshes]

    @classmethod
    def get_mesh_spacings(cls, grid, tol=None):
        subgrids = cls.get_mesh_subgrids(grid, tol=tol)
        if subgrids is None:
            return None
        if tol is None:
            tol = grid.dtype.itemsize // 2

        mesh_spacings = [np.unique(np.round(np.diff(g), tol)) for g in subgrids]
        if len(subgrids) == 1:
            mesh_spacings = mesh_spacings[0]
        return mesh_spacings

    @staticmethod
    def _is_meshgrid(grid):
        """
        Just checks if the shape of the grid
        is consistent with being a meshgrid
        :param grid:
        :type grid:
        :return:
        :rtype:
        """

        shape = grid.shape
        return (
            len(shape) > 1
            and any(shape[0] != s for s in shape[1:])
            and (shape[0] == len(shape) - 1)
        ) # might introduce edge-case...? Hard to know

    @classmethod
    def get_mesh_type(cls, grid, check_product_grid=True, check_regular_grid=True, tol=None):
        """
        Determines what kind of grid we're working with

        :param grid:
        :type grid: np.ndarray
        :return: mesh_type
        :rtype: MeshType
        """

        grid = np.asanyarray(grid).view(np.ndarray)

        if grid.dtype == np.dtype(object):
            return MeshType.Indeterminate

        ndim = grid.ndim
        shape = grid.shape
        cdim = ndim-1

        # we either need grid points or a full mesh
        if ndim > 2 and shape[-1] != cdim and shape[0] != cdim:
            raise ValueError("don't know how to interpret mesh with shape {}".format(shape))

        if ndim > 1 and shape[-1] != cdim and shape[0] == cdim:
            grid = np.moveaxis(grid, 0, ndim-1) # make full mesh

        if ndim == 1:
            mesh_spacings = cls.get_mesh_spacings(grid, tol=tol)
            if len(mesh_spacings) == 1:
                return MeshType.Regular
            else:
                return MeshType.Structured # all grids are structured in 1D
        elif ndim == 2 and shape[-1] != 1: # this means we got gridpoints
            if tol is None:
                tol = grid.dtype.itemsize // 2
            subgrids = cls.get_mesh_subgrids(grid)  # check if can be read as a structured grid
            if grid.shape[0] == np.prod([len(x) for x in subgrids]):
                # check if it really is a product grid
                if check_product_grid:
                    true_mesh = np.moveaxis(np.array(np.meshgrid(*subgrids, indexing='ij')), 0, ndim-1)
                    test_grid = np.reshape(grid, true_mesh.shape)
                    product_grid = np.allclose(true_mesh, test_grid, atol=10 ** (-tol))
                else:
                    product_grid = True

                if product_grid and check_regular_grid:
                    spacings = [np.unique(np.round(np.diff(x), tol)) for x in subgrids]
                    regular_grid = all(len(x) == 1 for x in spacings)
                else:
                    regular_grid = product_grid

                if regular_grid:
                    return MeshType.Regular
                elif product_grid:
                    return MeshType.Structured
                else:
                    return MeshType.Unstructured

            else:
                # should try to check for semistructured grids...
                return MeshType.Unstructured
        else:
            # means we _probably_ have a structured grid
            if tol is None:
                tol = grid.dtype.itemsize // 2
            subgrids = cls.get_mesh_subgrids(grid, tol=tol)
            # raise Exception(
            #     np.prod(grid.shape[:-1]), grid.shape,
            #     np.prod([len(x) for x in subgrids]), subgrids
            # )
            if np.prod(grid.shape[:-1]) == np.prod([len(x) for x in subgrids]): # check if it could be a product grid
                # check if it really is a product grid
                if check_product_grid:
                    true_mesh = np.moveaxis(np.array(np.meshgrid(*subgrids, indexing='ij')), 0, ndim-1)
                    product_grid = np.allclose(true_mesh, grid, atol=10**(-tol))
                else:
                    product_grid = True

                if product_grid and check_regular_grid:
                    spacings = [np.unique(np.round(np.diff(x), tol)) for x in subgrids]
                    regular_grid = all(len(x) == 1 for x in spacings)
                else:
                    regular_grid = product_grid

                if regular_grid:
                    return MeshType.Regular
                elif product_grid:
                    return MeshType.Structured
                else:
                    return MeshType.Unstructured
            else:
                # should try to check for semistructured grids...
                return MeshType.Unstructured

    @classmethod
    def RegularMesh(cls, *mesh_specs):
        """
        Builds a grid from multiple linspace arguments,
        basically insuring it's structured (if non-Empty)
        :param mesh_specs:
        :type mesh_specs:
        :return:
        :rtype:
        """
        # should probably handle the empty mesh subcase...
        coords = [ np.linspace(*m) for m in mesh_specs ]

        if any(len(x) == 0 for x in coords):
            raise ValueError("can't handle empty linspaces")
        cg = np.array(np.meshgrid(*coords, indexing='ij'))
        wat = np.moveaxis(cg, 0, len(mesh_specs))
        return cls(wat, mesh_type=MeshType.Regular)



"""
McUtils
A simple package for working with common McCoy group problems
"""

# pulled from Ryan's stuff, lightly modified
import glob

from setuptools import setup, find_packages

short_description = __doc__.split("\n")

try:
    with open("README.md", "r") as handle:
        long_description = handle.read()
except:
    long_description = "\n".join(short_description[2:])

def get_version():
    import subprocess
    run_out = subprocess.run(['git', 'describe', '--tags', '--abbrev=0'], capture_output=True)
    return run_out.stdout.decode().strip().strip("v")

setup(
    # Self-descriptive entries which should always be present
    name='mccoygroup-mcutils',
    author='Mark Boyer',
    author_email='b3m2a1@uw.edu',
    description=short_description[0],
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=get_version(),
    license='MIT',

    # Which Python importable modules should be included when your package is installed
    # Handled automatically by setuptools. Use 'exclude' to prevent some specific
    # subpackage(s) from being added, if needed
    packages=find_packages(),

    # Optional include package data to ship with your package
    # Customize MANIFEST.in if the general case does not suit your needs
    # Comment out this line to prevent the files from being packaged with your software
    # include_package_data=True

    install_requires=[
        "numpy>=1.8", # shouldn't need to restrict the python version anymore
        "scipy>=1.7.0",
        "h5py>=2.10.0",
        "numba>=0.57.1",
        "matplotlib>=3.3.4",
        # "jupyterlab>=3.3.2",
        # "ipywidgets>=7.6.3",
        # "nglview>=3.0.1",
        # "sympy>=1.9"
    ],

    include_package_data=True,
    # data_files=[
    #     # like `jupyter nbextension install --sys-prefix`
    #     ("share/jupyter/labextensions/ActiveHTMLWidget", [
    #         "my_fancy_module/static/index.js",
    #     ]),
    #     # like `jupyter nbextension enable --sys-prefix`
    #     ("etc/jupyter/nbconfig/notebook.d", [
    #         "jupyter-config/nbconfig/notebook.d/my_fancy_module.json"
    #     ]),
    #     # like `jupyter serverextension enable --sys-prefix`
    #     ("etc/jupyter/jupyter_notebook_config.d", [
    #         "jupyter-config/jupyter_notebook_config.d/my_fancy_module.json"
    #     ])
    # ],
    data_files=[
        ('share/jupyter/nbextensions/ActiveHTMLWidget', glob.glob('McUtils/Jupyter/JHTML/ActiveHTMLWidget/nbextension/*')),
        ('share/jupyter/labextensions/ActiveHTMLWidget/static', glob.glob('McUtils/Jupyter/JHTML/ActiveHTMLWidget/labextension/static/*.js')),
        ('share/jupyter/labextensions/ActiveHTMLWidget', glob.glob('McUtils/Jupyter/JHTML/ActiveHTMLWidget/labextension/*.json')),
        ('share/jupyter/labextensions/ActiveHTMLWidget', ['McUtils/Jupyter/JHTML/ActiveHTMLWidget/install.json']),
        ('etc/jupyter/nbconfig/notebook.d', ['McUtils/Jupyter/JHTML/ActiveHTMLWidget/ActiveHTMLWidget.json']),
    ]

)


import os, sys

lib_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, lib_dir)
try:
    from .ZachLib import *
except ImportError:
    from .src.setup import load
    if load() is None:
        raise
    from .ZachLib import *
finally:
    sys.path.pop(0)


"""
Provides a framework for using coordinates with explicit reference to an underlying coordinate system
"""


__all__ = []
from .CoordinateSystemConverter import *; from .CoordinateSystemConverter import __all__ as exposed
__all__ += exposed
from .CommonCoordinateSystems import *; from .CommonCoordinateSystems import __all__ as exposed
__all__ += exposed
from .CoordinateSystem import *; from .CoordinateSystem import __all__ as exposed
__all__ += exposed
from .CompositeCoordinateSystems import *; from .CompositeCoordinateSystems import __all__ as exposed
__all__ += exposed
from .GenericInternalCoordinateSystem import *; from .GenericInternalCoordinateSystem import __all__ as exposed
__all__ += exposed
from .CoordinateSet import *; from .CoordinateSet import __all__ as exposed
__all__ += exposed



import numpy as np

__all__ = [
    'infer_inds_dtype',
    'infer_int_dtype',
    'flatten_dtype',
    'unflatten_dtype',
    'recast_permutation',
    'recast_indices',
    'downcast_index_array',
    'numeric_types',
    'is_atomic',
    'is_numeric',
    'is_zero',
    'flatten_inds'
]

numeric_types = (int, float, np.integer, np.floating)
atomic_types = numeric_types + (str, )
def is_atomic(obj, types=None):
    if types is None: types = atomic_types
    return isinstance(obj, types) or (
            isinstance(obj, np.ndarray) and obj.shape == () and is_atomic(obj[()], types)
    )
def is_numeric(obj, types=None):
    if types is None: types = numeric_types
    return is_atomic(obj, types=types)
def is_zero(obj, numeric_types=None):
    return is_numeric(obj, types=numeric_types) and obj == 0

def downcast_index_array(a, max_val):
    return a.astype(infer_inds_dtype(max_val))

def recast_permutation(permutation_array):
    a = np.asanyarray(permutation_array)
    max_val = a.shape[-1]
    return downcast_index_array(a, max_val)

def recast_indices(indexing_array):
    a = np.asanyarray(indexing_array)
    max_val = np.max(a)
    return downcast_index_array(a, max_val)

def infer_inds_dtype(max_size):
    return np.min_scalar_type(max_size) # reset
    # needed the memory help back...
    # if max_size < 256:
    #     minimal_dtype = 'uint8'
    # elif max_size < 65535:
    #     minimal_dtype = 'uint16'
    # elif max_size < 4294967295:
    #     minimal_dtype = 'uint32'
    # else:
    #     minimal_dtype = 'uint64'
    # return minimal_dtype

def infer_int_dtype(max_dim):
    return np.min_scalar_type(-(max_dim+1))
    # max_dim = abs(max_dim)
    # if max_dim < 128:
    #     minimal_dtype = 'int8'
    # elif max_dim < 32768:
    #     minimal_dtype = 'int16'
    # elif max_dim < 2147483648:
    #     minimal_dtype = 'int32'
    # else:
    #     minimal_dtype = 'int64'
    # return minimal_dtype

def flatten_dtype(ar, dtype=None):
    """
    Extracted from the way NumPy treats unique
    Coerces ar into a compound dtype so that it can be treated
    like a 1D array for set operations
    """

    # Must reshape to a contiguous 2D array for this to work...
    orig_shape, orig_dtype = ar.shape, ar.dtype
    ar = ar.reshape(orig_shape[0], np.prod(orig_shape[1:], dtype=np.intp))
    ar = np.ascontiguousarray(ar)
    if dtype is None:
        dtype = [('f{i}'.format(i=i), ar.dtype) for i in range(ar.shape[1])]
    # At this point, `ar` has shape `(n, m)`, and `dtype` is a structured
    # data type with `m` fields where each field has the data type of `ar`.
    # In the following, we create the array `consolidated`, which has
    # shape `(n,)` with data type `dtype`.
    try:
        if ar.shape[1] > 0:
            consolidated = ar.view(dtype)
            if len(consolidated.shape) > 1:
                consolidated = consolidated.squeeze()
                if consolidated.shape == ():
                    consolidated = np.expand_dims(consolidated, 0)
        else:
            # If ar.shape[1] == 0, then dtype will be `np.dtype([])`, which is
            # a data type with itemsize 0, and the call `ar.view(dtype)` will
            # fail.  Instead, we'll use `np.empty` to explicitly create the
            # array with shape `(len(ar),)`.  Because `dtype` in this case has
            # itemsize 0, the total size of the result is still 0 bytes.
            consolidated = np.empty(len(ar), dtype=dtype)
    except TypeError:
        # There's no good way to do this for object arrays, etc...
        msg = 'The axis argument to `coerce_dtype` is not supported for dtype {dt}'
        raise TypeError(msg.format(dt=ar.dtype))

    return consolidated, dtype, orig_shape, orig_dtype

def unflatten_dtype(consolidated, orig_shape, orig_dtype, axis=None):
    """
    Converts a coerced array back to a full array
    :param consolidated:
    :type consolidated:
    :param orig_shape:
    :type orig_shape:
    :param orig_dtype:
    :type orig_dtype:
    :param axis: where to shift the main axis
    :type axis:
    :return:
    :rtype:
    """
    n = len(consolidated)
    uniq = consolidated.view(orig_dtype)
    uniq = uniq.reshape(n, *orig_shape[1:])
    if axis is not None:
        uniq = np.moveaxis(uniq, 0, axis)
    return uniq


def flatten_inds(A, *idx_blocks):
    idx_blocks = [
        (
            (i + A.ndim) if i < 0 else i,
            (j + A.ndim) if j < 0 else j
        )
        for i, j in idx_blocks
    ]
    block_map = {
        i: np.prod(A.shape[i:j+1], dtype=int)
        for i, j in idx_blocks
    }
    new_shape = [
        A.shape[k]
        if not any(i <= k and k <= j for i, j in idx_blocks) else
        block_map.get(k)
        for k in range(A.ndim)
    ]
    new_shape = [s for s in new_shape if s is not None]

    return A.reshape(new_shape)


"""
Provides a CoordinateSet class that acts as a symbolic extension of np.ndarray to provide an explicit basis
"""
import numpy as np
from .CoordinateSystem import CoordinateSystem, CoordinateSystemError
from .CommonCoordinateSystems import CartesianCoordinates3D
from .CoordinateUtils import is_multiconfig

__all__ = [
    "CoordinateSet"
]

__reload_hook__ = ['.CoordinateSystem', '.CommonCoordinateSystems', '.CoordinateUtils']

######################################################################################################
##
##                                   CoordinateSet Class
##
######################################################################################################
class CoordinateSet(np.ndarray):
    """
    A subclass of np.ndarray that lives in an explicit coordinate system and can convert between them
    """
    # note that the subclassing procedure here is very non-standard because of how ndarray works

    def __new__(cls, coords, system=CartesianCoordinates3D, converter_options=None):
        self = np.asarray(coords).view(cls)
        self.system = system
        self.converter_options = converter_options
        return self

    def __init__(self, coords, system = CartesianCoordinates3D, converter_options = None):
        # all the heavy lifting is done in _validate
        # and we only want to ever call this once
        self._validate()

    def __array_finalize__(self, coords):
        # basically just a validator...
        if coords is None:
            return None

        self.system = getattr(coords, "system", CartesianCoordinates3D)
        self.converter_options = getattr(coords, "converter_options", None)

    def _validate(self):
        base_dim = self.system.dimension
        if base_dim is not None:
            if isinstance(base_dim, int):
                core_dim = self.shape[-1]
            else:
                cdim = self.shape[-len(base_dim):]
                base_dim = tuple(base_dim)
                core_dim = tuple( a if a is None else b for a, b in zip(base_dim, cdim) )
            if base_dim != core_dim:
                raise CoordinateSystemError(
                    "Dimension of basis {} '{}' and dimension of coordinate set '{}' misaligned".format(
                        self.system.name,
                        self.system.dimension,
                        core_dim
                    )
                )

    def __str__(self):
        return "{}({}, {})".format(type(self).__name__, self.system.name, super().__str__())

    def __eq__(self, other):
        if isinstance(other, CoordinateSet) and self.system is not other.system:
            return False
        else:
            return super().__eq__(other)

    @property
    def multiconfig(self):
        """Determines whether self.coords represents multiple configurations of the coordinates

        :return:
        :rtype:
        """
        return is_multiconfig(self)

    def transform(self, tf):
        """Applies a transformation to the stored coordinates

        :param tf: the transformation function to apply to the system
        :type tf:
        :return:
        :rtype:
        """
        coords = self
        new_coords = tf(coords)
        return type(self)(new_coords, system=self.system)

    def convert(self, system, **kw):
        """Converts across coordinate systems

        :param system: the target coordinate system
        :type system: CoordinateSystem
        :return: new_coords
        :rtype: CoordinateSet
        """
        cops = self.converter_options
        if cops is None:
            cops = {}
        kw = dict(cops, **kw)
        res = self.system.convert_coords(self, system, **kw)
        if not isinstance(res, np.ndarray):
            new_coords, ops = res
        else:
            new_coords = res
            ops = {}
        return type(self)(new_coords, system=system, converter_options=ops)

    def derivatives(self,
                    function,
                    order=1,
                    coordinates=None,
                    result_shape=None,
                    **fd_options
                    ):
        """
        Takes derivatives of `function` with respect to the current geometry

        :param function:
        :type function:
        :param order:
        :type order:
        :param coordinates:
        :type coordinates:
        :param fd_options:
        :type fd_options:
        :return:
        :rtype:
        """
        return self.system.derivatives(self, function, order=order,
                                       coordinates=coordinates, result_shape=result_shape, **fd_options)

    def jacobian(self,
                 system,
                 order=1,
                 coordinates=None,
                 converter_options=None,
                 all_numerical=False,
                 analytic_deriv_order=None,
                 **fd_options
                 ):
        """
        Delegates to the jacobian function of the current coordinate system.


        :param system:
        :type system:
        :param order:
        :type order:
        :param mesh_spacing:
        :type mesh_spacing:
        :param prep:
        :type prep:
        :param coordinates:
        :type coordinates:
        :param fd_options:
        :type fd_options:
        :return:
        :rtype:
        """

        cops = self.converter_options
        if cops is None:
            cops = {}
        if converter_options is None:
            converter_options = {}
        kw = dict(cops, **converter_options)
        if self.system.coordinate_shape is None:
            self.system.coordinate_shape = self.shape
        return self.system.jacobian(self,
                                    system,
                                    order=order,
                                    coordinates=coordinates,
                                    converter_options=kw,
                                    all_numerical=all_numerical,
                                    analytic_deriv_order=analytic_deriv_order,
                                    **fd_options
                                    )




import os, shutil

from .DocWalker import DocWalker
from ..Formatters.TemplateEngine import ResourceLocator

__all__ = [
    "DocBuilder"
]

class DocBuilder:
    """
    A documentation builder class that uses a `DocWalker`
    to build documentation, but which also has support for more
    involved use cases, like setting up a `_config.yml` or other
    documentation template things.


    :related: .DocWalker.DocWalker, .DocWalker.ModuleWriter, .DocWalker.ClassWriter,
              .DocWalker.FunctionWriter, .DocWalker.MethodWriter, .DocWalker.ObjectWriter, .DocWalker.IndexWriter
    """

    defaults_root = os.path.dirname(__file__)
    default_config_file="_config.yml"

    def __init__(self,
                 packages=None,
                 config=None,
                 target=None,
                 root=None,
                 config_file=None,
                 templates_directory=None,
                 examples_directory=None,
                 tests_directory=None,
                 readme=None
                 ):
        """
        :param packages: list of package configs to write
        :type packages: Iterable[str|dict]
        :param config: parameters for _config.yml file
        :type config: dict
        :param target: target directory to which files should be written
        :type target: str
        :param root: root directory
        :type root: str
        :param root: root directory
        :type root: str
        """
        self.packages = packages
        self.config = config
        self.target = target
        self.root = root
        self.template_dir=self.get_template_locator(templates_directory, use_repo_templates='gh_repo' in self.config)
        self.examples_dir=examples_directory
        self.tests_directory=tests_directory
        self.config_file=self.default_config_file if config_file is None else config_file

        if isinstance(readme, str):
            if os.path.isfile(readme):
                with open(readme) as r:
                    readme = r.read()
        self.readme=readme

    default_template_extension = 'templates'
    default_repo_extension = 'repo_templates'
    def get_template_locator(self, template_directory, use_repo_templates=False):
        if not isinstance(template_directory, ResourceLocator):
            template_extension = [self.default_template_extension]
            if use_repo_templates:
                template_extension = [self.default_repo_extension] + template_extension
            tdirs = [os.path.dirname(os.path.abspath(__file__))]
            if template_directory is not None:
                if isinstance(template_directory, str):
                    template_directory = [template_directory]
                template_directory = list(template_directory)
                tdirs = list(template_directory) + tdirs
                template_directory = ResourceLocator([template_directory, (tdirs, template_extension)])
            else:
                template_directory = ResourceLocator([(tdirs, template_extension)])
        return template_directory

    config_defaults = {
        "theme":"McCoyGroup/finx",
        "gh_username":"McCoyGroup-bot",
        "footer": ""
    }
    def load_config(self):
        """
        Loads the config file to be used and fills in template parameters

        :return:
        :rtype:
        """
        cfg = self.config_defaults.copy()
        if self.config is not None:
            cfg.update(**self.config)
        self.config = cfg

        test_cfg = self.template_dir.locate(self.config_file)
        cfg_file = self.config_file if os.path.isfile(self.config_file) else (
            test_cfg if os.path.isfile(test_cfg) else self.template_dir.locate('_config.yml')
        )

        if os.path.isfile(cfg_file):
            print('reading config template from {}'.format(cfg_file))
            with open(cfg_file) as config_dump:
                cfg_string = config_dump.read().format(**cfg)

            return cfg_string

    def create_layout(self):
        """
        Creates the documentation layout that will be expanded upon by
        a `DocWalker`

        :return:
        :rtype:
        """

        if self.config is not None:
            conf = self.load_config()
            if isinstance(conf, str):
                config_file = os.path.join(self.target, '_config.yml') # hard coded for now, can change later
                print('writing config file to {}'.format(config_file))
                try:
                    os.makedirs(os.path.dirname(config_file))
                except OSError:
                    pass
                with open(config_file, 'w') as out:
                    out.write(conf)

        for fname in ['404.html', 'Contributing.md', 'Gemfile']:
            conf_file = os.path.join(self.target, fname)  # hard coded for now, can change later
            base_file = self.template_dir.locate(fname)
            if not os.path.isfile(conf_file) and os.path.isfile(fname):
                print('writing {} file to {}'.format(fname, fname))
                try:
                    os.makedirs(os.path.dirname(conf_file))
                except OSError:
                    pass
                shutil.copy(base_file, conf_file)

        # do other layout stuff maybe in the future?

    def load_walker(self):
        """
        Loads the `DocWalker` used to write docs.
        A hook that can be overriden to sub in different walkers.

        :return:
        :rtype:
        """
        return DocWalker(
            out=self.target,
            description=self.readme,
            template_locator=self.template_dir,
            examples_directory=self.examples_dir,
            tests_directory=self.tests_directory,
            **self.config
        )

    def build(self):
        """
        Writes documentation layout to `self.target`

        :return:
        :rtype:
        """
        print("using templates from {}".format(self.template_dir))
        print("using examples from {}".format(self.examples_dir))
        self.create_layout()
        walker = self.load_walker()
        return walker.write(self.packages)



"""
Basic layer for Schema validation that provides a superset of JSON schema validation
"""

__all__ = ["Schema"]

class Schema:
    """
    An object that represents a schema that can be used to test
    if an object matches that schema or not
    """

    def __init__(self, schema, optional_schema=None):
        self.schema = self.canonicalize_schema(schema)
        self.sub_schema = self.canonicalize_schema(optional_schema)

    @classmethod
    def canonicalize_schema(cls, schema):
        if schema is None:
            return None
        try:
            k0 = schema[0] # list-like duck typing
        except TypeError:
            pass
        else:
            schema = {k:None for k in schema}
        return schema

    def _validate_entry(self, obj, k, v, prop_getter, throw=False):
        match = True
        missing = False
        mistyped = False
        try:
            t = prop_getter(k)
        except (KeyError, AttributeError):
            missing = True
            match = False
            t = None
        except (TypeError,):
            mistyped = True
            match = False
            t = None
        else:
            if v is None:
                pass
            elif isinstance(v, (type, tuple)):
                match = isinstance(t, v)
            elif isinstance(v, Schema):
                match = v.validate(t, throw=throw)
            else:
                match = v(t)

        if not match and throw:
            if missing:
                raise KeyError("object {} doesn't match schema {}; key {} is missing".format(
                    obj, self, k
                ))
            elif mistyped:
                raise KeyError("object {} doesn't match schema {}; doesn't support attributes".format(
                    obj, self
                ))
            else:
                raise KeyError("object {} doesn't match schema {}; value {} doesn't match schema type {}".format(
                    obj, self, t, v
                ))
        return (t, match)

    def validate(self, obj, throw=True):
        """
        Validates that `obj` matches the provided schema
        and throws an error if not

        :param obj:
        :type obj:
        :param throw:
        :type throw:
        :return:
        :rtype:
        """

        prop_getter = ( lambda k,obj=obj:getattr(obj, k) ) if not hasattr(obj, '__getitem__') else obj.__getitem__
        for k,v in self.schema.items():
            if not self._validate_entry(obj, k, v, prop_getter, throw=throw)[1]:
                return False
        else:
            return True

    def to_dict(self, obj, throw=True):
        """
        Converts `obj` into a plain `dict` representation

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        res = {}
        prop_getter = ( lambda k,obj=obj:getattr(obj, k) ) if not hasattr(obj, '__getitem__') else obj.__getitem__
        for k, v in self.schema.items():
            t, m = self._validate_entry(obj, k, v, prop_getter, throw=throw)
            if not m:
                return None
            else:
                res[k] = t
        if self.sub_schema is not None:
            for k, v in self.sub_schema.items():
                t, m = self._validate_entry(obj, k, v, prop_getter, throw=False)
                if m:
                    res[k] = t
        return res

    def __repr__(self):
        return "{}({}, {})".format(type(self).__name__, self.schema, self.sub_schema)


"""
Defines classes for providing different approaches to fitting.
For the most part, the idea is to use `scipy.optimize` to do the actual fitting process,
but we layer on conveniences w.r.t. specification of bases and automation of the actual fitting process
"""

import numpy as np, scipy.optimize as opt, enum
from collections import OrderedDict as odict

__all__ = [
    "FittableModel",
    "LinearFittableModel",
    "LinearFitBasis"
]

class FittableModel:
    """
    Defines a model that can be fit
    """
    def __init__(self, parameters, function, pre_fit = False, covariance = None):
        if not isinstance(parameters, odict):
            ...
        elif isinstance(parameters, np.ndarray) or isinstance(parameters[0], (int, float, np.floating, np.integer)):
            param_names = ["c"+str(i) for i in range(len(parameters))]
            parameters = odict(zip(param_names, parameters))
        else:
            parameters = odict(parameters)
        self.initial_parameters = parameters
        self._param_names = list(parameters.keys())
        if pre_fit:
            self._parameter_values = np.array(parameters.values())
        else:
            self._parameter_values = None
        self.covariance = covariance
        self.func = function

    @property
    def parameters(self):
        if not self.fitted:
            raise ValueError("{}: model hasn't been fitted".format(type(self).__name__))
        return odict(zip(self._param_names, self._parameter_values))
    @property
    def parameter_values(self):
        return self._parameter_values
    @property
    def parameter_names(self):
        return self._param_names
    @property
    def fitted(self):
        return self._parameter_values is not None

    def fit(self, xdata, ydata=None, fitter=None, **methopts):
        """
        Fits the model to the data using scipy.optimize.curve_fit or a function that provides the same interface

        :param points:
        :type points:
        :param methopts:
        :type methopts:
        :return:
        :rtype:
        """
        if fitter is None:
            fitter = opt.curve_fit

        if self.fitted:
            p0 = self._parameter_values
        else:
            p0 = np.array(self.initial_parameters.values())

        if 'p0' not in methopts:
            methopts['p0'] = p0

        if ydata is None:
            ydata = xdata[:, -1]
            xdata = xdata[:, :-1]
        params, cov = fitter(xdata, ydata, **methopts)
        self._parameter_values = params
        self.covariance = cov

    def get_parameter(self, name):
        """
        Returns the fitted value of the parameter given by 'name'
        :param name:
        :type name:
        :return:
        :rtype:
        """
        try:
            ind = self._param_names.index(name)
        except IndexError:
            ind = None
        if ind is None:
            raise KeyError("{}: model has no parameter '{}'".format(type(self).__name__, name))
        return self._parameter_values[ind]
    def __getitem__(self, item):
        return self.get_parameter(item)

    def evaluate(self, xdata):
        if not self.fitted:
            raise ValueError("{}: model hasn't been fitted".format(type(self).__name__))
        return self.func(xdata, *self._parameter_values)
    def __call__(self, xdata):
        return self.evaluate(xdata)

class LinearFittableModel(FittableModel):
    """
    Defines a class of models that can be expressed as linear expansions of basis functions.
    We _could_ define an alternate fit function by explicitly building & fitting a design matrix, but I think we're good on that for now
    """

    def __init__(self, basis, initial_params=None, pre_fit=False, covariance=None):
        if isinstance(basis, LinearFitBasis):
            names = basis.names
            funcs = basis.functions
            if initial_params is None:
                initial_params = odict((n, 1) for n in names)
        if initial_params is None:
            initial_params = [1]*len(basis)
        self.basis = basis
        func = lambda xdata, *coefs, basis=basis: np.dot(np.array(coefs), np.array([b(xdata) for b in basis]))
        super().__init__(initial_params, func, pre_fit=pre_fit, covariance=covariance)

    def evaluate(self, xdata):
        if not self.fitted:
            raise ValueError("{}: model hasn't been fitted".format(type(self).__name__))
        return np.dot(self.parameter_values, np.array([b(xdata) for b in self.basis]))

class LinearFitBasis:
    """
    Provides a container to build bases of functions for fitting.
    Asks for a generator for each dimension, which is just a function that takes an integer and returns a basis function at that order.
    Product functions are taken up to some max order
    """
    def __init__(self, *generators, order=3):
        """

        :param generators: the generating functions for the bases in each dimenion
        :type generators: Iterable[function]
        :param order: the maximum order for the basis functions (currently turning off coupling isn't possible, but that could come)
        :type order: int
        """
        self.ndim = len(generators)
        self.generators = generators
        self._order = order
        self._functions = None
        self._basis_names = None

    @property
    def functions(self):
        if self._functions is None:
            self._functions, self._basis_names = self.construct_basis()
        return self._functions
    @property
    def names(self):
        if self._basis_names is None:
            self._functions, self._basis_names = self.construct_basis()
        return self._basis_names
    @property
    def order(self):
        return self._order
    @order.setter
    def order(self, order):
        if not isinstance(order, int):
            raise TypeError("{}: basis order must be an int, got '{}'".format(
                type(self).__name__,
                order
            ))
        self._functions = None
        self._order = order
    def construct_basis(self):
        import itertools as ip

        if self.ndim > 1:
            inds = ip.product(range(self._order), repeat=self.ndim)
            funcs = []
            names = []
            for i in inds:
                # filter to only pick up basis functions where the
                if sum(i) <= self._order:
                    prods = tuple(f(n) for f,n in zip(self.generators, i))
                    # build basis function by taking products
                    funcs.append(lambda xdata, fs=prods: np.prod([f(xdata[..., n]) for n, f in enumerate(fs)]))
                    names.append("c"+"".join(str(n) for n in i))
        else:
            funcs = [self.generators[0](n) for n in range(self._order)]
            names = ["c"+str(n) for n in range(self._order)]

        return funcs, names

    # just here for convenience
    fourier_series = lambda x,k: np.cos(k*x)
    power_series = lambda x,n: x**n



